import math
import torch
from torchdt import DType
from torchdt.ops.misc_ops import (
    DTBroadcastToFunction,
    DTCloneFunction,
    DTSqueezeFunction,
    DTUnsqueezeFunction,
    DTStackFunction,
    DTCatFunction,
    DTChunkFunction,
    DTWhereFunction,
    DTPadFunction,
    DTGetItemFunction,
    DTSetItemFunction,
    DTToFunction,
    DTViewFunction,
    DTContiguousFunction,
    DTRepeatFunction,
    DTFlattenFunction,
    DTReshapeFunction,
)

@DType.register_func(torch.broadcast_to, torch.Tensor.expand,
                     cast=("input",))
def dt_broadcast_to(input, shape):
    return DTBroadcastToFunction.apply(input, shape)

@DType.register_func(torch.clone, torch.Tensor.clone,
                     cast=("input",))
def dt_clone(input, memory_format=torch.preserve_format):
    return DTCloneFunction.apply(input, memory_format)

@DType.register_func(torch.squeeze, torch.Tensor.squeeze,
                     cast=("input",))
def dt_squeeze(input, dim=None):
    return DTSqueezeFunction.apply(input, dim)

@DType.register_func(torch.unsqueeze, torch.Tensor.unsqueeze,
                     cast=("input",))
def dt_unsqueeze(input, dim):
    return DTUnsqueezeFunction.apply(input, dim)

@DType.register_func(torch.stack,
                     cast=("tensors",))
def dt_stack(tensors, dim=0, *, out=None):
    result = DTStackFunction.apply(dim, *tensors)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.cat,
                     cast=("tensors",))
def dt_cat(tensors, dim=0, *, out=None):
    result = DTCatFunction.apply(dim, *tensors)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.chunk, torch.Tensor.chunk,
                     cast=("input",))
def dt_chunk(input, chunks, dim=0):
    return DTChunkFunction.apply(input, chunks, dim)

@DType.register_func(torch.where, torch.Tensor.where,
                     cast=("input", "other"))
def dt_where(condition, input, other, *, out=None):
    result = DTWhereFunction.apply(condition, input, other)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.nn.functional.pad,
                     cast=("input", "value"))
def dt_pad(input, pad, mode="constant", value=0):
    return DTPadFunction.apply(input, pad, mode, value)

@DType.register_func(torch.Tensor.__getitem__,
                     cast=("input",))
def dt_getitem(input, index):
    return DTGetItemFunction.apply(input, index)

# turns an index into a set of index tensors for each dimension
def _make_index_tensors(index, shape):
    flat_count = math.prod(shape)
    labels = torch.arange(flat_count).reshape(shape)
    flat_idx = labels[index].reshape(-1)
    return torch.unravel_index(flat_idx, shape)

@DType.register_func(torch.Tensor.__setitem__,
                     cast=("input", "value"))
def dt_setitem(input, index, value):
    index = _make_index_tensors(index, input.shape)
    return DTSetItemFunction.apply(input, index, value)

@DType.register_func(torch.Tensor.to,
                     cast=("input",))
def dt_to(input, device=None):
    return DTToFunction.apply(input, device)

@DType.register_func(torch.Tensor.view,
                     cast=("input",))
def dt_view(input, *shape):
    return DTViewFunction.apply(input, shape)

@DType.register_func(torch.Tensor.contiguous,
                     cast=("input",))
def dt_contiguous(input, memory_format=torch.preserve_format):
    return DTContiguousFunction.apply(input, memory_format)

@DType.register_func(torch.Tensor.repeat,
                     cast=("input",))
def dt_repeat(input, *repeats):
    return DTRepeatFunction.apply(input, repeats)

@DType.register_func(torch.flatten, torch.Tensor.flatten,
                     cast=("input",))
def dt_flatten(input, start_dim=0, end_dim=-1):
    return DTFlattenFunction.apply(input, start_dim, end_dim)

@DType.register_func(torch.reshape, torch.Tensor.reshape,
                     cast=("input",))
def dt_reshape(input, shape):
    return DTReshapeFunction.apply(input, shape)

@DType.register_func(torch.Tensor.item,
                     cast=("input",))
def dt_item(input):
    return input.to_float().item()


import torch
import torch.optim.lr_scheduler as lr_sched
from torchdt.optim import DTOptimizer
from collections import Counter
from bisect import bisect_right
import warnings

__all__ = [
    "StepLR",
    "MultiStepLR",
    "ReduceLROnPlateau",
]

def _param_groups_val_list(optimizer, key):
    return [
        group[key].clone() if isinstance(group[key], torch.Tensor) else group[key]
        for group in optimizer.param_groups
    ]

def _update_param_group_val(param_group, key, val):
    if isinstance(param_group[key], torch.Tensor):
        param_group[key].copy_(val)
    else:
        param_group[key] = val

def _warn_get_lr_called_within_step(lr_scheduler):
    if not lr_scheduler._get_lr_called_within_step:
        warnings.warn(
            "To get the last learning rate computed by the scheduler, "
            "please use `get_last_lr()`.",
            UserWarning,
            stacklevel=2,
        )


class StepLR(lr_sched.LRScheduler):

    def __init__(
        self,
        optimizer,
        step_size,
        gamma = 0.1,
        last_epoch = -1,
    ):
        self.step_size = step_size
        self.gamma = gamma
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        _warn_get_lr_called_within_step(self)

        if (self.last_epoch == 0) or (self.last_epoch % self.step_size != 0):
            return _param_groups_val_list(self.optimizer, "lr")
        return [
            group["lr"] * torch.tensor(self.gamma, device=group["lr"].device)
            for group in self.optimizer.param_groups
        ]

    def _get_closed_form_lr(self):
        return [
            base_lr * torch.tensor(
                self.gamma ** (self.last_epoch // self.step_size),
                device=base_lr.device
            )
            for base_lr in self.base_lrs
        ]

class MultiStepLR(lr_sched.LRScheduler):

    def __init__(
        self,
        optimizer,
        milestones,
        gamma = 0.1,
        last_epoch = -1,
    ):
        self.milestones = Counter(milestones)
        self.gamma = gamma
        super().__init__(optimizer, last_epoch)

    def get_lr(self):
        _warn_get_lr_called_within_step(self)

        if self.last_epoch not in self.milestones:
            return _param_groups_val_list(self.optimizer, "lr")
        return [
            group["lr"] * torch.tensor(
                self.gamma ** self.milestones[self.last_epoch],
                device=group["lr"].device,
            )
            for group in self.optimizer.param_groups
        ]

    def _get_closed_form_lr(self):
        milestones = sorted(self.milestones.elements())
        return [
            base_lr * torch.tensor(
                self.gamma ** bisect_right(milestones, self.last_epoch),
                device=base_lr.device,
            )
            for base_lr in self.base_lrs
        ]

class ReduceLROnPlateau(lr_sched.LRScheduler):

    def __init__(
        self,
        optimizer,
        mode = "min",
        factor = 0.1,
        patience = 10,
        threshold = 1e-4,
        threshold_mode = "rel",
        cooldown = 0,
        min_lr = 0,
        eps = 1e-8,
    ):
        if factor >= 1.0:
            raise ValueError("Factor should be < 1.0.")
        self.factor = factor

        if not isinstance(optimizer, DTOptimizer):
            raise TypeError(f"{type(optimizer).__name__} is not a DTOptimizer")
        self.optimizer = optimizer

        if isinstance(min_lr, (list, tuple)):
            if len(min_lr) != len(optimizer.param_groups):
                raise ValueError(
                    f"expected {len(optimizer.param_groups)} min_lrs, got {len(min_lr)}"
                )
            self.default_min_lr = None
            self.min_lrs = list(min_lr)

        else:
            self.default_min_lr = min_lr
            self.min_lrs = [min_lr] * len(optimizer.param_groups)

        self.patience = patience
        self.cooldown = cooldown
        self.eps = eps
        self.last_epoch = 0
        self._last_lr = _param_groups_val_list(self.optimizer, "lr")
        self._init_is_better(
            mode=mode, threshold=threshold, threshold_mode=threshold_mode
        )
        self._reset()

    def _reset(self):
        self.best = None
        self.cooldown_counter = 0
        self.num_bad_epochs = 0

    def step(self, metrics):
        current = float(metrics)
        self.last_epoch += 1

        if self._is_better(current, self.best):
            self.best = current
            self.num_bad_epochs = 0
        else:
            self.num_bad_epochs += 1

        if self.in_cooldown:
            self.cooldown_counter -= 1
            self.num_bad_epochs = 0

        if self.num_bad_epochs > self.patience:
            self._reduce_lr(self.last_epoch)
            self.cooldown_counter = self.cooldown
            self.num_bad_epochs = 0

        self._last_lr = _param_groups_val_list(self.optimizer, "lr")

    def _reduce_lr(self, epoch):
        if len(self.optimizer.param_groups) != len(self.min_lrs):
            if self.default_min_lr is None:
                raise RuntimeError(
                    "The number of param groups in the `optimizer` "
                    f"({len(self.optimizer.param_groups)}) differs "
                    f"from when `ReduceLROnPlateau` was initialized "
                    f"({len(self.min_lrs)}), usually due to a new "
                    "param group being added to the optimizer. Please "
                    "modify the `min_lrs` field to match the length "
                    "of the `optimizer` param groups."
                )
            else:
                self.min_lrs = [self.default_min_lr] * len(self.optimizer.param_groups)

        for i, param_group in enumerate(self.optimizer.param_groups):
            old_lr = param_group["lr"]
            factor = torch.tensor(self.factor, device=old_lr.device)
            min_lr = torch.tensor(self.min_lrs[i], device=old_lr.device)
            eps = torch.tensor(self.eps, device=old_lr.device)
            new_lr = torch.maximum(old_lr * factor, min_lr)
            if old_lr - new_lr > eps:
                _update_param_group_val(param_group, "lr", new_lr)

    def _is_better(self, a, best):
        if best is None:
            return True

        if self.mode == "min" and self.threshold_mode == "rel":
            rel_epsilon = 1.0 - self.threshold
            return a < best * rel_epsilon

        elif self.mode == "min" and self.threshold_mode == "abs":
            return a < best - self.threshold

        elif self.mode == "max" and self.threshold_mode == "rel":
            rel_epsilon = self.threshold + 1.0
            return a > best * rel_epsilon

        else:  # mode == 'max' and epsilon_mode == 'abs':
            return a > best + self.threshold

    def _init_is_better(self, mode, threshold, threshold_mode) -> None:
        if mode not in {"min", "max"}:
            raise ValueError("mode " + mode + " is unknown!")
        if threshold_mode not in {"rel", "abs"}:
            raise ValueError("threshold mode " + threshold_mode + " is unknown!")

        self.mode = mode
        self.threshold = threshold
        self.threshold_mode = threshold_mode


import torch
from torchdt import DType

@DType.register_func(torch.nn.init.uniform_, torch.Tensor.uniform_)
def dt_uniform_(tensor, a=0.0, b=1.0, *, generator=None):
    result = torch.nn.init.uniform_(torch.empty(tensor.size(), device=tensor.device), a=a, b=b, generator=generator)
    return tensor.data.copy_(tensor.__class__(result))

@DType.register_func(torch.nn.init.normal_, torch.Tensor.normal_)
def dt_normal_(tensor, mean=0.0, std=1.0, *, generator=None):
    result = torch.nn.init.normal_(torch.empty(tensor.size(), device=tensor.device), mean=mean, std=std, generator=generator)
    return tensor.data.copy_(tensor.__class__(result))

@DType.register_func(torch.nn.init.constant_, torch.Tensor.fill_)
def dt_constant_(tensor, val):
    # here we can precompute the result since it's constant
    val_dt = tensor.__class__(val, device=tensor.device)
    return tensor.data.copy_(val_dt)

@DType.register_func(torch.nn.init.ones_)
def dt_ones_(tensor):
    # precompute the one value in the target dtype
    one_dt = tensor.__class__(1.0, device=tensor.device)
    return tensor.data.copy_(one_dt)

@DType.register_func(torch.nn.init.zeros_, torch.Tensor.zero_)
def dt_zeros_(tensor):
    # precompute the zero value in the target dtype
    zero_dt = tensor.__class__(0.0, device=tensor.device)
    return tensor.data.copy_(zero_dt)

@DType.register_func(torch.nn.init.xavier_uniform_)
def dt_xavier_uniform_(tensor, gain=1.0, generator=None):
    result = torch.nn.init.xavier_uniform_(torch.empty(tensor.size(), device=tensor.device), gain=gain, generator=generator)
    return tensor.data.copy_(tensor.__class__(result))

@DType.register_func(torch.nn.init.xavier_normal_)
def dt_xavier_normal_(tensor, gain=1.0, generator=None):
    result = torch.nn.init.xavier_normal_(torch.empty(tensor.size(), device=tensor.device), gain=gain, generator=generator)
    return tensor.data.copy_(tensor.__class__(result))

@DType.register_func(torch.nn.init.kaiming_uniform_)
def dt_kaiming_uniform_(tensor, a=0.0, mode='fan_in', nonlinearity='leaky_relu', generator=None):
    result = torch.nn.init.kaiming_uniform_(torch.empty(tensor.size(), device=tensor.device), a=a, mode=mode, nonlinearity=nonlinearity, generator=generator)
    return tensor.data.copy_(tensor.__class__(result))

@DType.register_func(torch.nn.init.kaiming_normal_)
def dt_kaiming_normal_(tensor, a=0.0, mode='fan_in', nonlinearity='leaky_relu', generator=None):
    result = torch.nn.init.kaiming_normal_(torch.empty(tensor.size(), device=tensor.device), a=a, mode=mode, nonlinearity=nonlinearity, generator=generator)
    return tensor.data.copy_(tensor.__class__(result))


import torch
from torch import Tensor, CharTensor, ShortTensor, IntTensor, LongTensor
from typing import Union, Callable

InternalTensor = Union[CharTensor, ShortTensor, IntTensor, LongTensor]

def register_op(dtype_cls: type, method: str) -> Callable:
    """Decorator to register an operation for a given DType subclass."""
    def decorator(func: Callable) -> Callable:
        ops_cls = dtype_cls.ops
        if not hasattr(ops_cls, method):
            raise ValueError(f"{ops_cls.__name__} has no method '{method}' to register.")
        setattr(ops_cls, method, classmethod(func))
        return func
    return decorator

def register_base_op(method: str) -> Callable:
    """Decorator to register a base operation."""
    def decorator(func: Callable) -> Callable:
        if not hasattr(OpsBase, method):
            raise ValueError(f"OpsBase has no method '{method}' to register.")
        setattr(OpsBase, method, classmethod(func))
        return func
    return decorator

class OpsBase:

    # ========== Useful helper functions ==========

    @classmethod
    def scalar_from_float(cls, x: Union[float, int]) -> InternalTensor:
        x_tensor = torch.tensor(x, dtype=torch.float32)
        return cls.from_float(x_tensor)

    @classmethod
    def zeros(cls, size, device=None):
        return torch.full(size, cls.scalar_from_float(0.0), dtype=cls.dtype.int_dtype, device=device)

    @classmethod
    def zeros_like(cls, x):
        return torch.full_like(x, cls.scalar_from_float(0.0), dtype=cls.dtype.int_dtype, device=x.device)

    @classmethod
    def ones(cls, size, device=None):
        return torch.full(size, cls.scalar_from_float(1.0), dtype=cls.dtype.int_dtype, device=device)

    @classmethod
    def ones_like(cls, x):
        return torch.full_like(x, cls.scalar_from_float(1.0), dtype=cls.dtype.int_dtype, device=x.device)

    @classmethod
    def full(cls, size, fill_value, device=None):
        return torch.full(size, cls.scalar_from_float(fill_value), dtype=cls.dtype.int_dtype, device=device)

    @classmethod
    def full_like(cls, x, fill_value):
        return torch.full_like(x, cls.scalar_from_float(fill_value), dtype=cls.dtype.int_dtype)

    @classmethod
    def sum_to_size(cls, x: InternalTensor, target_size: torch.Size) -> InternalTensor:
        if list(x.shape) == list(target_size):
            return x

        x_shape = list(x.shape)
        target_shape = list(target_size)
        if x.dim() > len(target_shape):
            target_shape = [1] * (x.dim() - len(target_shape)) + target_shape

        # reduce dimensions that were broadcasted
        leading = x.dim() - len(target_shape)
        if leading > 0:
            x = cls.sum(x, dim=tuple(range(leading)), keepdim=False)
            x_shape = x_shape[leading:]

        # reduce dimensions where target size is 1 but tensor has a larger size
        reduce_dims = [i for i, (ts, gs) in enumerate(zip(x_shape, target_shape)) if gs == 1 and ts != 1]
        if reduce_dims:
            x = cls.sum(x, dim=tuple(reduce_dims), keepdim=True)

        return x.reshape(target_size)

    # ========== Operations to be implemented by subclasses ==========

    @classmethod
    def from_float(cls, x: Tensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def to_float(cls, x: InternalTensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def add(cls, x: InternalTensor, y: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def sub(cls, x: InternalTensor, y: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def mul(cls, x: InternalTensor, y: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def div(cls, x: InternalTensor, y: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def pow(cls, x: InternalTensor, y: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def sign(cls, x: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def ge(cls, x: InternalTensor, y: InternalTensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def gt(cls, x: InternalTensor, y: InternalTensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def le(cls, x: InternalTensor, y: InternalTensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def lt(cls, x: InternalTensor, y: InternalTensor) -> Tensor:
        raise NotImplementedError

    # ========== Backward operations for c++ ops ==========

    @classmethod
    def matmul_backward(cls, grad_output: InternalTensor, A: InternalTensor,
                        B: InternalTensor) -> tuple[InternalTensor, InternalTensor]:
        raise NotImplementedError

    @classmethod
    def conv2d_backward(cls, grad_output: InternalTensor, input: InternalTensor,
                        weight: InternalTensor, stride: tuple[int], padding: tuple[int],
                        dilation: tuple[int], has_bias: bool, groups: int
                        ) -> tuple[InternalTensor, InternalTensor, InternalTensor]:
        raise NotImplemented

    # ========== 'Base' operations with default implementations ==========

    @classmethod
    def sum(cls, x: InternalTensor, dim=None, keepdim=False) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def square(cls, x: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def sqrt(cls, x: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def reciprocal(cls, x: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def exp(cls, x: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def log(cls, x: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def prod(cls, x: InternalTensor, dim=None, keepdim=False) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def mean(cls, x: InternalTensor, dim=None, keepdim=False) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def var(cls, x: InternalTensor, correction: InternalTensor, dim=None, keepdim=False) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def matmul(cls, A: InternalTensor, B: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def transpose(cls, A: InternalTensor, dim0: int, dim1: int) -> InternalTensor:
        raise NotImplementedError


    @classmethod
    def neg(cls, x: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def abs(cls, x: InternalTensor) -> InternalTensor:
        raise NotImplementedError


    @classmethod
    def equal(cls, x: InternalTensor, y: InternalTensor) -> bool:
        raise NotImplementedError

    @classmethod
    def eq(cls, x: InternalTensor, y: InternalTensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def ne(cls, x: InternalTensor, y: InternalTensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def isclose(cls, x: InternalTensor, y: InternalTensor,
                rtol: InternalTensor, atol: InternalTensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def allclose(cls, x: InternalTensor, y: InternalTensor,
                 rtol: InternalTensor, atol: InternalTensor) -> Tensor:
        raise NotImplementedError

    @classmethod
    def any(cls, x: InternalTensor, dim=None, keepdim=False) -> Tensor:
        raise NotImplementedError

    @classmethod
    def all(cls, x: InternalTensor, dim=None, keepdim=False) -> Tensor:
        raise NotImplementedError

    @classmethod
    def isin(cls, x: InternalTensor, y: InternalTensor,
             assume_unique = False, invert = False) -> Tensor:
        raise NotImplementedError

    @classmethod
    def maximum(cls, x: InternalTensor, y: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def minimum(cls, x: InternalTensor, y: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def max(cls, x: InternalTensor, dim=None, keepdim=False) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def min(cls, x: InternalTensor, dim=None, keepdim=False) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def argmax(cls, x: InternalTensor, dim=None, keepdim=False) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def argmin(cls, x: InternalTensor, dim=None, keepdim=False) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def clamp(cls, x: InternalTensor, min: InternalTensor = None, max: InternalTensor = None) -> InternalTensor:
        raise NotImplementedError


    @classmethod
    def relu(cls, x: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def leaky_relu(cls, x: InternalTensor, negative_slope: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def threshold(cls, x: InternalTensor, threshold: InternalTensor, value: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def tanh(cls, x: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def sigmoid(cls, x: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def logsigmoid(cls, x: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def softmin(cls, x: InternalTensor, dim: int = None) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def softmax(cls, x: InternalTensor, dim: int = None) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def log_softmax(cls, x: InternalTensor, dim: int = None) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def hardtanh(cls, x: InternalTensor, min_val: InternalTensor, max_val: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def glu(cls, x: InternalTensor, dim: int = -1) -> InternalTensor:
        raise NotImplementedError


    @classmethod
    def broadcast_to(cls, x: InternalTensor, shape) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def clone(cls, x: InternalTensor, memory_format = torch.preserve_format) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def squeeze(cls, x: InternalTensor, dim=None) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def unsqueeze(cls, x: InternalTensor, dim: int) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def stack(cls, tensors: list[InternalTensor], dim: int = 0) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def cat(cls, tensors: list[InternalTensor], dim: int = 0) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def chunk(cls, x: InternalTensor, chunks: int, dim: int = 0) -> list[InternalTensor]:
        raise NotImplementedError

    @classmethod
    def where(cls, condition: Tensor, x: InternalTensor, y: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def pad(cls, x: InternalTensor, pad: tuple[int], mode: str = 'constant', value: InternalTensor = None):
        raise NotImplementedError

    @classmethod
    def getitem(cls, x: InternalTensor, index) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def setitem(cls, x: InternalTensor, index, value: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def to(cls, x: InternalTensor, device) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def view(cls, x: InternalTensor, shape) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def contiguous(cls, x: InternalTensor) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def repeat(cls, x: InternalTensor, repeats: tuple[int]) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def flatten(cls, x: InternalTensor, start_dim: int = 0, end_dim: int = -1) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def reshape(cls, x: InternalTensor, shape: tuple[int]) -> InternalTensor:
        raise NotImplementedError


    @classmethod
    def mse_loss(cls, x: InternalTensor, y: InternalTensor,
                 reduction: str = 'mean', weight: InternalTensor = None) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def l1_loss(cls, x: InternalTensor, y: InternalTensor,
                 reduction: str = 'mean', weight: InternalTensor = None) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def binary_cross_entropy(cls, x: InternalTensor, y: InternalTensor,
                             weight: InternalTensor = None, reduction: str = 'mean') -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def binary_cross_entropy_with_logits(cls, x: InternalTensor, y: InternalTensor,
                                         weight: InternalTensor = None, reduction: str = 'mean',
                                         pos_weight: InternalTensor = None) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def nll_loss(cls, x: InternalTensor, y: InternalTensor,
                 weight: InternalTensor = None, reduction: str = 'mean',
                 ignore_index: int = -100) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def poisson_nll_loss(cls, x: InternalTensor, y: InternalTensor,
                         eps: InternalTensor, log_input: bool = True,
                         full: bool = False, reduction: str = 'mean') -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def hinge_embedding_loss(cls, x: InternalTensor, y: InternalTensor,
                             margin: InternalTensor = None, reduction: str = 'mean') -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def kl_div(cls, x: InternalTensor, y: InternalTensor,
                reduction: str = 'mean', log_target: bool = False) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def margin_ranking_loss(cls, x1: InternalTensor, x2: InternalTensor,
                            y: InternalTensor, margin: InternalTensor,
                            reduction: str = 'mean') -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def gaussian_nll_loss(cls, x: InternalTensor, y: InternalTensor,
                          var: InternalTensor, eps: InternalTensor,
                          full: bool = False, reduction: str = 'mean') -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def huber_loss(cls, x: InternalTensor, y: InternalTensor,
                   delta: InternalTensor, reduction: str = 'mean',
                   weight: InternalTensor = None) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def smooth_l1_loss(cls, x: InternalTensor, y: InternalTensor,
                       beta: InternalTensor, reduction: str = 'mean') -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def cross_entropy(cls, x: InternalTensor, y: InternalTensor,
                      weight: InternalTensor = None, ignore_index: int = -100,
                      reduction: str = 'mean', label_smoothing: InternalTensor = None) -> InternalTensor:
        raise NotImplementedError


    @classmethod
    def linear(cls, x: InternalTensor, weight: InternalTensor, bias: InternalTensor = None) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def dropout(cls, x: InternalTensor, p: float = 0.5) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def conv2d(cls, x: InternalTensor, weight: InternalTensor,
               bias: InternalTensor = None, stride: tuple[int] = 1,
               padding: tuple[int] = 0, dilation: tuple[int] = 1,
               groups: int = 1) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def avg_pool2d(cls, x: InternalTensor, kernel_size: tuple[int],
                  stride: tuple[int] = None, padding: tuple[int] = 0,
                  ceil_mode: bool = False, count_include_pad: bool = True,
                  divisor_override=None) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def adaptive_avg_pool2d(cls, x: InternalTensor, output_size: tuple[int]) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def max_pool2d(cls, x: InternalTensor, kernel_size: tuple[int],
                  stride: tuple[int] = None, padding: tuple[int] = 0,
                  dilation: tuple[int] = 1, ceil_mode: bool = False,
                  return_indices: bool = False) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def batch_norm(cls, x: InternalTensor, running_mean: InternalTensor,
                   running_var: InternalTensor, momentum: InternalTensor,
                   eps: InternalTensor, weight: InternalTensor = None,
                   bias: InternalTensor = None, training: bool = False) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def layer_norm(cls, x: InternalTensor, eps: InternalTensor,
                   normalized_shape: tuple[int], weight: InternalTensor = None,
                   bias: InternalTensor = None) -> InternalTensor:
        raise NotImplementedError


    @classmethod
    def triton_sgd_step(cls, param: InternalTensor, grad: InternalTensor,
                     momentum_buffer: InternalTensor, lr: InternalTensor,
                     momentum: InternalTensor, dampening: InternalTensor,
                     weight_decay: InternalTensor, nesterov: bool,
                     maximize: bool) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def triton_madam_step(cls, params: InternalTensor, grad: InternalTensor,
                        exp_avg_sq: InternalTensor, lr: InternalTensor,
                        beta: InternalTensor, eps: InternalTensor, 
                        g_bound: InternalTensor, max: InternalTensor,
                        bias_corr: InternalTensor, use_pow: bool,
                        maximize: bool) -> InternalTensor:
        raise NotImplementedError

    @classmethod
    def triton_adam_step(cls, params: InternalTensor, grad: InternalTensor,
                        exp_avg: InternalTensor, exp_avg_sq: InternalTensor,
                        max_exp_avg_sq: InternalTensor, lr: InternalTensor,
                        beta1: InternalTensor, beta2: InternalTensor,
                        eps: InternalTensor, weight_decay: InternalTensor,
                        bias_corr1: InternalTensor, bias_corr2: InternalTensor,
                        amsgrad: bool, maximize: bool) -> tuple[InternalTensor, InternalTensor, InternalTensor]:
        raise NotImplementedError


import torch
from torch import Tensor
from typing import Any, Optional, Union, Type, Dict, Callable, Tuple
import functools
import inspect

from torchdt.transforms import register_collate_dtype_fn
from torchdt.ops import OpsBase, register_op, register_cpp_ops

_float_dtype = {
    8: torch.float8_e5m2, # we have several variants to pick from
    16: torch.float16,
    32: torch.float32,
    64: torch.float64,
}

_int_dtype = {
    8: torch.uint8,
    16: torch.int16,
    32: torch.int32,
    64: torch.int64
}

# for functions that should not be overridden by __torch_function__
no_override_funcs = {
    Tensor.backward,
    Tensor.copy_,
    Tensor.detach,
    Tensor.dim,
    Tensor.numel,
    Tensor.requires_grad_,
    Tensor.register_hook,
    Tensor.register_post_accumulate_grad_hook,
    Tensor.size,
    Tensor.data_ptr,
    Tensor.__reduce_ex__
}
# for functions that should not be overridden by __torch_function__
# where it is hard to reference them, so we do it by name
no_override_func_names = {
    "__get__",
}

class GradAccumHook:

    def __init__(self, tensor, dtype):
        self.value = dtype(torch.zeros(tensor.size(), device=tensor.device), requires_grad=False)
        self.dtype = dtype

        self.grad_hook_handle = tensor.register_hook(self.grad_hook)
        if tensor.is_leaf:
            self.grad_accum_hook_handle = tensor.register_post_accumulate_grad_hook(self.accumulate_hook)

    def grad_hook(self, grad):
        if grad is None:
            return None
        return self.value

    def accumulate_hook(self, tensor):
        tensor.grad.copy_(self.value)

    def register_edge_hook(self, edge, arg_index):

        def edge_hook(grad_inputs, grad_outputs):
            if grad_inputs[arg_index] is not None:
                # __torch_function__ doesn't work inside hooks, so we must
                # re-enable it manually with a context manager.
                with torch._C._EnableTorchFunction():
                    self.value = self.value + grad_inputs[arg_index].as_subclass(self.dtype)

        edge.node.register_hook(edge_hook)

    # def __del__(self):
    #     if hasattr(self, "grad_hook_handle"):
    #         self.grad_hook_handle.remove()
    #     if hasattr(self, "grad_accum_hook_handle"):
    #         self.grad_accum_hook_handle.remove()

class DType(Tensor):
    """
    Parent class for custom dtypes (posit, LNS, etc) that live in a Tensor
    but expose their own semantics.
    """
    bitwidth: int = 32 # subclasses override
    torch_funcs: Dict[Callable, Callable] = {} # mapping from 'torch.' function to custom implementation

    def __new__(
            cls,
            data: Any,
            *,
            internal: bool = False,
            device: Optional[Union[str, torch.device]] = None,
            requires_grad: Optional[bool] = None,
            memory_format: torch.memory_format = torch.preserve_format,
    ):
        if isinstance(data, DType):
            if data.__class__ == cls:
                payload = data
            else:
                payload = data.to_float()
                payload = ToDType.apply(payload, cls)
        elif isinstance(data, torch.Tensor):
            if internal:
                if data.dtype != cls.float_dtype:
                    payload = data.view(cls.float_dtype)
                else:
                    payload = data
            else:
                payload = data.to(dtype=torch.float32, device=device, memory_format=memory_format)
                payload = ToDType.apply(payload, cls)
        else:
            if internal:
                payload = torch.tensor(data, dtype=cls.int_dtype, device=device).view(cls.float_dtype)
            else:
                payload = torch.tensor(data, dtype=torch.float32, device=device)
            payload = ToDType.apply(payload, cls)
            payload = payload.to(memory_format=memory_format)

        obj = payload.as_subclass(cls)
        if requires_grad is None:
            if isinstance(data, torch.Tensor) and data.requires_grad:
                obj.requires_grad_(True)
        else:
            obj.requires_grad_(requires_grad)
        return obj

    def __init_subclass__(cls, bitwidth: int = 32, cpp_backend=None, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.float_dtype = _float_dtype[bitwidth]
        cls.int_dtype = _int_dtype[bitwidth]

        if bitwidth not in _float_dtype:
            raise ValueError(
                f"{cls.__name__} has invalid bitwidth {bitwidth}. "
                f"Must be one of {tuple(_float_dtype.keys())}."
            )
        cls.bitwidth = bitwidth
        cls.cpp_backend = cpp_backend

        if cls is DType:
            return # don't register base class

        # tell the collate function to handle this DType
        # this is used for DataLoader batching
        register_collate_dtype_fn(cls)

        # tell torch that this DType is safe to save/load
        torch.serialization.add_safe_globals([cls])

        # create a subclass of Ops for this DType
        ops_name = f"{cls.__name__}Ops"
        namespace = {
            '__module__': OpsBase.__module__,
            'dtype': cls,
        }
        ops_cls = type(ops_name, (OpsBase,), namespace)
        cls.ops = ops_cls

        # allow normal imports to see it
        # module = sys.modules[cls.__module__]
        # setattr(module, ops_name, ops_cls)

    @classmethod
    def enable_cpp_backend(cls, backend=None):
        if cls.cpp_backend is None and backend is None:
            raise ValueError(f"{cls.__name__} has no C++ backend to enable.")
        register_cpp_ops(cls, backend or cls.cpp_backend)

    def _track_operation(self, edge, arg_index):
        """
        Registers a hook to track the operation that produced this DType tensor.
        This is used to accumulate gradients from different paths in the computation
        graph since PyTorch will internally add these, but since they are DType
        tensors we must perform our custom DType addition.
        """
        self._grad_accum_hook.register_edge_hook(edge, arg_index)

    def requires_grad_(self, requires_grad: bool = True):
        """Sets the requires_grad flag for this DType tensor in-place."""
        super().requires_grad_(requires_grad)

        if requires_grad:
            self._grad_accum_hook = GradAccumHook(self, self.__class__)

        return self

    def backward(self, gradient=None, retain_graph=None, create_graph=False, inputs=None):
        """
        Computes the gradient of current DType tensor with respect to the graph leaves.
        This method is analogous to the standard PyTorch `Tensor.backward()` method, but
        works with DType tensors. See

        https://docs.pytorch.org/docs/stable/generated/torch.Tensor.backward.html

        for more details. Note that the `gradient` parameter, if provided, will be converted
        to the same DType as `self` before being used in the backward pass.
        """

        if gradient is None:

            if self.numel() != 1:
                raise RuntimeError("grad can be implicitly created only for scalar outputs")

            # create a tensor of ones in the same dtype as self
            gradient = self.__class__(torch.ones(self.size()), device=self.device)

        elif gradient.__class__ != self.__class__:
            gradient = self.__class__(gradient, device=self.device, requires_grad=False)

        # manually set the incoming gradients for the output
        # tensor since no hooks will be registered for it.
        self._grad_accum_hook.value.copy_(gradient)

        return super().backward(
            gradient=gradient,
            retain_graph=retain_graph,
            create_graph=create_graph,
            inputs=inputs
        )

    @property
    def grad(self):
        """The gradient of this DType tensor."""
        if super().grad is None:
            return None
        return super().grad.as_subclass(self.__class__)

    @classmethod
    def register_func(
        cls,
        *torch_funcs: Callable,
        cast: Tuple[Union[str, int]] = ()):
        """Decorator to register a custom implementation for a torch.* function."""

        def decorator(func: Callable) -> Callable:
            sig = inspect.signature(func)
            param_names = list(sig.parameters.keys())

            def id_to_name(identifier: Union[str, int]):
                if isinstance(identifier, int):
                    try:
                        return param_names[identifier]
                    except IndexError:
                        raise IndexError(f"positional index {identifier} out of range for {func.__name__}")
                return identifier # assume string

            cast_names = [id_to_name(x) for x in cast]

            @functools.wraps(func)
            def wrapped_func(*args, _dtype_cls=None, **kwargs):
                if _dtype_cls is None:
                    raise ValueError("_dtype_cls must be provided when calling registered torch function.")

                bound = sig.bind_partial(*args, **kwargs)
                for pname in cast_names:
                    if pname in bound.arguments:
                        if bound.arguments[pname] is not None and type(bound.arguments[pname]) != _dtype_cls:
                            # convert argument to the correct DType subclass - can also handle lists, tuples, etc?
                            if isinstance(bound.arguments[pname], (list, tuple)):
                                bound.arguments[pname] = type(bound.arguments[pname])(
                                    _dtype_cls(x) for x in bound.arguments[pname]
                                )
                            else:
                                bound.arguments[pname] = _dtype_cls(bound.arguments[pname])

                return func(*bound.args, **bound.kwargs)

            for torch_func in torch_funcs:
                cls.torch_funcs[torch_func] = wrapped_func

            return wrapped_func

        return decorator

    @classmethod
    def __torch_function__(cls, func, types, args=..., kwargs=None):
        """Override to handle torch.* functions for this DType subclass."""

        if kwargs is None:
            kwargs = {}

        if func not in cls.torch_funcs:
            if func in no_override_funcs or func.__name__ in no_override_func_names:
                return super().__torch_function__(func, types, args, kwargs)
            raise NotImplementedError(f"{cls.__name__} has no implementation for torch function '{func.__name__}'.")

        # pass cls to cast any floating point or number arguments to tensors of this DType
        return cls.torch_funcs[func](*args, _dtype_cls=cls, **kwargs)

    @classmethod
    def register_op(cls, method: str):
        """Decorator to register an operation for this DType subclass."""
        return register_op(cls, method)

    @property
    def _float(self) -> Tensor:
        "Return the underlying storage as a plain *float* tensor."
        return self.as_subclass(Tensor)

    @property
    def _int(self) -> Tensor:
        "Integer bit-view of the same storage (no copy)."
        return self._float.view(_int_dtype[self.bitwidth])

    def to_float(self):
        return self.ops.to_float(self._int)

    def __float__(self):
        return self.ops.to_float(self._int).item()

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self.to_float()}, bitwidth={self.bitwidth}, "
            f"shape={tuple(self.shape)}, device={self.device})"
        )

class ToDType(torch.autograd.Function):

    @staticmethod
    def forward(ctx, input: Tensor, dtype: Type[DType]) -> DType:
        return dtype.ops.from_float(input).view(dtype.float_dtype)

    @staticmethod
    def backward(ctx, grad_output: DType) -> Tensor:
        return grad_output.to_float(), None


import torch
from torchdt.autograd import DTFunction
from torchdt.ops import register_base_op

class DTSignFunction(DTFunction):

    @staticmethod
    def forward(ops, x):
        return ops.sign(x)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        pass

    @staticmethod
    def backward(ctx, ops, grad_output):
        grad_x = ops.zeros_like(grad_output)
        return grad_x

@register_base_op("neg")
def dt_neg(ops, x):
    return ops.mul(ops.scalar_from_float(-1.0), x)

class DTNegFunction(DTFunction):

    @staticmethod
    def forward(ops, x):
        return ops.neg(x)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        pass

    @staticmethod
    def backward(ctx, ops, grad_output):
        grad_x = ops.neg(grad_output)
        return grad_x

@register_base_op("abs")
def dt_abs(ops, x):
    return ops.mul(ops.sign(x), x)

class DTAbsFunction(DTFunction):

    @staticmethod
    def forward(ops, x):
        return ops.abs(x)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, = inputs
        ctx.save_for_backward(x)

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, = ctx.saved_tensors

        x_less_than_zero = ops.lt(x, ops.scalar_from_float(0.0))
        grad_x = torch.where(x_less_than_zero, ops.neg(grad_output), grad_output)

        return grad_x


import torch
from torchdt.autograd import DTFunction, DTNonDifferentiableFunction
from torchdt.ops import register_base_op

@register_base_op("equal")
def dt_equal(ops, x, y):
    return torch.equal(x, y)

class DTEqualFunction(DTNonDifferentiableFunction):

    output_indices = []

    @staticmethod
    def forward(ops, x, y):
        return ops.equal(x, y)

@register_base_op("eq")
def dt_eq(ops, x, y):
    return torch.eq(x, y)

class DTEqFunction(DTNonDifferentiableFunction):

    output_indices = []

    @staticmethod
    def forward(ops, x, y):
        return ops.eq(x, y)

@register_base_op("ne")
def dt_ne(ops, x, y):
    return torch.ne(x, y)

class DTNeFunction(DTNonDifferentiableFunction):

    output_indices = []

    @staticmethod
    def forward(ops, x, y):
        return ops.ne(x, y)

class DTGeFunction(DTNonDifferentiableFunction):

    output_indices = []

    @staticmethod
    def forward(ops, x, y):
        return ops.ge(x, y)

class DTGtFunction(DTNonDifferentiableFunction):

    output_indices = []

    @staticmethod
    def forward(ops, x, y):
        return ops.gt(x, y)

class DTLeFunction(DTNonDifferentiableFunction):

    output_indices = []

    @staticmethod
    def forward(ops, x, y):
        return ops.le(x, y)

class DTLtFunction(DTNonDifferentiableFunction):

    output_indices = []

    @staticmethod
    def forward(ops, x, y):
        return ops.lt(x, y)

@register_base_op("isclose")
def dt_isclose(ops, x, y, rtol, atol):
    abs_diff = ops.abs(ops.sub(x, y))
    eps = ops.add(atol, ops.mul(rtol, ops.abs(y)))
    return ops.le(abs_diff, eps)

class DTIscloseFunction(DTNonDifferentiableFunction):

    output_indices = []

    @staticmethod
    def forward(ops, x, y, rtol, atol):
        return ops.isclose(x, y, rtol, atol)

@register_base_op("allclose")
def dt_allclose(ops, x, y, rtol, atol):
    is_close = ops.isclose(x, y, rtol, atol)
    return torch.all(is_close)

class DTAllcloseFunction(DTNonDifferentiableFunction):

    output_indices = []

    @staticmethod
    def forward(ops, x, y, rtol, atol):
        return ops.allclose(x, y, rtol, atol)

@register_base_op("any")
def dt_any(ops, x, dim=None, keepdim=False):
    return torch.any(ops.ne(x, ops.scalar_from_float(0.0)), dim=dim, keepdim=keepdim)

class DTAnyFunction(DTNonDifferentiableFunction):

    output_indices = []

    @staticmethod
    def forward(ops, x, dim=None, keepdim=False):
        return ops.any(x, dim=dim, keepdim=keepdim)

@register_base_op("all")
def dt_all(ops, x, dim=None, keepdim=False):
    return torch.all(ops.ne(x, ops.scalar_from_float(0.0)), dim=dim, keepdim=keepdim)

class DTAllFunction(DTNonDifferentiableFunction):

    output_indices = []

    @staticmethod
    def forward(ops, x, dim=None, keepdim=False):
        return ops.all(x, dim=dim, keepdim=keepdim)

@register_base_op("isin")
def dt_isin(ops, x, y, assume_unique=False, invert=False):
    return torch.isin(x, y, assume_unique=assume_unique, invert=invert)

class DTIsinFunction(DTNonDifferentiableFunction):

    output_indices = []

    @staticmethod
    def forward(ops, x, y, assume_unique=False, invert=False):
        return ops.isin(x, y, assume_unique=assume_unique, invert=invert)

@register_base_op("maximum")
def dt_maximum(ops, x, y):
    x_greater_y = ops.gt(x, y)
    return ops.where(x_greater_y, x, y)

class DTMaximumFunction(DTFunction):

    @staticmethod
    def forward(ops, x, y):
        return ops.maximum(x, y)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, y = inputs
        ctx.save_for_backward(x, y)

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, y = ctx.saved_tensors

        x_y_equal = ops.eq(x, y)
        half_grad_output = ops.mul(grad_output, ops.scalar_from_float(0.5))

        grad_x = torch.where(x_y_equal, half_grad_output, torch.where(
            ops.gt(x, y), grad_output, ops.scalar_from_float(0.0)
        ))
        grad_y = torch.where(x_y_equal, half_grad_output, torch.where(
            ops.gt(y, x), grad_output, ops.scalar_from_float(0.0)
        ))

        grad_x = ops.sum_to_size(grad_x, x.shape)
        grad_y = ops.sum_to_size(grad_y, y.shape)

        return grad_x, grad_y

@register_base_op("minimum")
def dt_minimum(ops, x, y):
    x_less_y = ops.lt(x, y)
    return ops.where(x_less_y, x, y)

class DTMinimumFunction(DTFunction):

    @staticmethod
    def forward(ops, x, y):
        return ops.minimum(ops, x, y)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, y = inputs
        ctx.save_for_backward(x, y)

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, y = ctx.saved_tensors

        x_y_equal = ops.eq(x, y)
        half_grad_output = ops.mul(grad_output, ops.scalar_from_float(0.5))

        grad_x = torch.where(x_y_equal, half_grad_output, torch.where(
            ops.lt(x, y), grad_output, ops.scalar_from_float(0.0)
        ))
        grad_y = torch.where(x_y_equal, half_grad_output, torch.where(
            ops.lt(y, x), grad_output, ops.scalar_from_float(0.0)
        ))

        grad_x = ops.sum_to_size(grad_x, x.shape)
        grad_y = ops.sum_to_size(grad_y, y.shape)

        return grad_x, grad_y

# todo: use tree (pairwise) reduction for sufficiently large dims
@register_base_op("max")
def dt_max(ops, x, dim=None, keepdim=False):
    return_indices = dim is not None

    if dim is None:
        x = x.view(-1)
        dim = 0

    # Initial running maximum (slice 0 along the reduction dimension)
    out = x.select(dim, 0).clone()

    # if we need to return indices
    if return_indices:
        max_idx = torch.zeros_like(out, dtype=torch.long)

    # iterate once over the dimension we are reducing
    for i in range(1, x.size(dim)):
        candidate = x.select(dim, i)
        mask = ops.gt(candidate, out) # where candidate > current_max
        out = torch.where(mask, candidate, out)
        if return_indices:
            max_idx = torch.where(mask, torch.full_like(max_idx, i), max_idx)

    if keepdim:
        out = out.unsqueeze(dim)
        if return_indices:
            max_idx = max_idx.unsqueeze(dim)

    if return_indices:
        return torch.return_types.max((out, max_idx))
    return out

class DTMaxFunction(DTFunction):

    output_indices = [0]

    @staticmethod
    def forward(ops, x, dim=None, keepdim=False):
        return ops.max(x, dim, keepdim)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, dim, keepdim = inputs
        if dim is None:
            ctx.save_for_backward(x, output)
        else:
            _, indices = output
            ctx.save_for_backward(x, indices)
        ctx.dim = dim
        ctx.keepdim = keepdim

    @staticmethod
    def backward(ctx, ops, grad_output, grad_indices=None): # grad_indices is not used
        if ctx.dim is None:
            x, result = ctx.saved_tensors

            max_values = torch.eq(x, result)
            grad_x = ops.div(grad_output, ops.scalar_from_float(max_values.sum()))

            return torch.where(max_values, grad_x, ops.scalar_from_float(0.0)), None, None, None

        x, indices = ctx.saved_tensors

        grad_x = ops.zeros_like(x)
        if ctx.keepdim:
            idx_expanded  = indices
            grad_expanded = grad_output
        else:
            idx_expanded  = indices.unsqueeze(ctx.dim)
            grad_expanded = grad_output.unsqueeze(ctx.dim)

        grad_x.scatter_(ctx.dim,
                        idx_expanded.expand(x.shape),
                        grad_expanded.expand(x.shape))

        return grad_x, None, None

# todo: use tree (pairwise) reduction for sufficiently large dims
@register_base_op("min")
def dt_min(ops, x, dim=None, keepdim=False):
    return_indices = dim is not None

    if dim is None:
        x = x.view(-1)
        dim = 0

    # Initial running minimum (slice 0 along the reduction dimension)
    out = x.select(dim, 0).clone()

    # if we need to return indices
    if return_indices:
        min_idx = torch.zeros_like(out, dtype=torch.long)

    # iterate once over the dimension we are reducing
    for i in range(1, x.size(dim)):
        candidate = x.select(dim, i)
        mask = ops.lt(candidate, out) # where candidate < current_min
        out = torch.where(mask, candidate, out)
        if return_indices:
            min_idx = torch.where(mask, torch.full_like(min_idx, i), min_idx)

    if keepdim:
        out = out.unsqueeze(dim)
        if return_indices:
            min_idx = min_idx.unsqueeze(dim)

    if return_indices:
        return torch.return_types.min((out, min_idx))
    return out

class DTMinFunction(DTFunction):

    output_indices = [0]

    @staticmethod
    def forward(ops, x, dim=None, keepdim=False):
        return ops.min(x, dim, keepdim)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, dim, keepdim = inputs
        if dim is None:
            ctx.save_for_backward(x, output)
        else:
            _, indices = output
            ctx.save_for_backward(x, indices)
        ctx.dim = dim
        ctx.keepdim = keepdim

    @staticmethod
    def backward(ctx, ops, grad_output, grad_indices=None): # grad_indices is not used
        if ctx.dim is None:
            x, result = ctx.saved_tensors

            min_values = torch.eq(x, result)
            grad_x = ops.div(grad_output, ops.scalar_from_float(min_values.sum()))

            return torch.where(min_values, grad_x, ops.scalar_from_float(0.0)), None, None, None

        x, indices = ctx.saved_tensors

        grad_x = ops.zeros_like(x)
        if ctx.keepdim:
            idx_expanded  = indices
            grad_expanded = grad_output
        else:
            idx_expanded  = indices.unsqueeze(ctx.dim)
            grad_expanded = grad_output.unsqueeze(ctx.dim)

        grad_x.scatter_(ctx.dim,
                        idx_expanded.expand(x.shape),
                        grad_expanded.expand(x.shape))

        return grad_x, None, None

@register_base_op("argmax")
def dt_argmax(ops, x, dim=None, keepdim=False):
    if dim is None:
        x = x.view(-1)
        dim = 0

    elif dim < 0:
        dim += x.dim()

    max_val = x.select(dim, 0).clone()
    max_idx = torch.zeros_like(max_val, dtype=torch.long)

    for i in range(1, x.size(dim)):
        candidate = x.select(dim, i)
        mask = ops.gt(candidate, max_val) # candidate > current best
        max_val = torch.where(mask, candidate, max_val)
        max_idx = torch.where(mask, torch.full_like(max_idx, i), max_idx)

    if keepdim:
        max_idx = max_idx.unsqueeze(dim)

    return max_idx

class DTArgmaxFunction(DTNonDifferentiableFunction):

    output_indices = []

    @staticmethod
    def forward(ops, x, dim=None, keepdim=False):
        return ops.argmax(x, dim, keepdim)

@register_base_op("argmin")
def dt_argmin(ops, x, dim=None, keepdim=False):
    if dim is None:
        x = x.view(-1)
        dim = 0

    elif dim < 0:
        dim += x.dim()

    min_val = x.select(dim, 0).clone()
    min_idx = torch.zeros_like(min_val, dtype=torch.long)

    for i in range(1, x.size(dim)):
        candidate = x.select(dim, i)
        mask = ops.lt(candidate, min_val) # candidate < current best
        min_val = torch.where(mask, candidate, min_val)
        min_idx = torch.where(mask, torch.full_like(min_idx, i), min_idx)

    if keepdim:
        min_idx = min_idx.unsqueeze(dim)

    return min_idx

class DTArgminFunction(DTNonDifferentiableFunction):

    output_indices = []

    @staticmethod
    def forward(ops, x, dim=None, keepdim=False):
        return ops.argmin(x, dim, keepdim)

@register_base_op("clamp")
def dt_clamp(ops, x, min=None, max=None):
    result = x.clone()

    if min is not None:
        lt_mask = ops.lt(result, min)
        result = torch.where(lt_mask, min, result)

    if max is not None:
        gt_mask = ops.gt(result, max)
        result = torch.where(gt_mask, max, result)

    return result

class DTClampFunction(DTFunction):

    @staticmethod
    def forward(ops, x, min=None, max=None):
        return ops.clamp(x, min, max)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, min, max = inputs
        ctx.save_for_backward(x, min, max)

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, min, max = ctx.saved_tensors

        grad_x = grad_output.clone()
        if min is not None:
            lt_mask = ops.lt(x, min)
            grad_x = torch.where(lt_mask, ops.scalar_from_float(0.0), grad_x)

        if max is not None:
            gt_mask = ops.gt(x, max)
            grad_x = torch.where(gt_mask, ops.scalar_from_float(0.0), grad_x)

        return grad_x, None, None


import torch
from torchdt import DType
from torchdt.ops.arithmetic_ops import (
    DTAddFunction,
    DTSubFunction,
    DTSumFunction,
    DTMulFunction,
    DTDivFunction,
    DTPowFunction,
    DTSquareFunction,
    DTSqrtFunction,
    DTReciprocalFunction,
    DTExpFunction,
    DTLogFunction,
    DTProdFunction,
    DTMeanFunction,
    DTVarFunction,
    DTMatmulFunction,
    DTTransposeFunction,
)

@DType.register_func(torch.add, torch.Tensor.add,
                     cast=("input", "other"))
def dt_add(input, other, *, alpha=1, out=None):
    if alpha != 1:
        other = torch.mul(other, alpha)
    result = DTAddFunction.apply(input, other)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.Tensor.__radd__,
                     cast=("self", "other"))
def dt_radd(self, other):
    return torch.add(other, self)

@DType.register_func(torch.sub, torch.Tensor.sub,
                     cast=("input", "other"))
def dt_sub(input, other, *, alpha=1, out=None):
    if alpha != 1:
        other = torch.mul(other, alpha)
    result = DTSubFunction.apply(input, other)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.Tensor.__rsub__,
                     cast=("self", "other"))
def dt_rsub(self, other):
    return torch.sub(other, self)

@DType.register_func(torch.sum, torch.Tensor.sum,
                     cast=("input",))
def dt_sum(input, dim=None, keepdim=False, *, out=None):
    result = DTSumFunction.apply(input, dim, keepdim)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.mul, torch.Tensor.mul,
                     cast=("input", "other"))
def dt_mul(input, other, *, out=None):
    result = DTMulFunction.apply(input, other)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.Tensor.__rmul__,
                     cast=("self", "other"))
def dt_rmul(self, other):
    return torch.mul(other, self)

@DType.register_func(torch.div, torch.Tensor.div,
                     cast=("input", "other"))
def dt_div(input, other, *, out=None):
    result = DTDivFunction.apply(input, other)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.Tensor.__rdiv__,
                     cast=("self", "other"))
def dt_rdiv(self, other):
    return torch.div(other, self)

@DType.register_func(torch.pow, torch.Tensor.pow, torch.Tensor.__pow__,
                     cast=("input", "exponent"))
def dt_pow(input, exponent, *, out=None):
    result = DTPowFunction.apply(input, exponent)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.Tensor.__rpow__,
                     cast=("self", "other"))
def dt_rpow(self, other):
    return torch.pow(other, self)

@DType.register_func(torch.square, torch.Tensor.square,
                     cast=("input"))
def dt_square(input, *, out=None):
    result = DTSquareFunction.apply(input)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.sqrt, torch.Tensor.sqrt,
                     cast=("input",))
def dt_sqrt(input, *, out=None):
    result = DTSqrtFunction.apply(input)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.reciprocal, torch.Tensor.reciprocal,
                     cast=("input",))
def dt_reciprocal(input, *, out=None):
    result = DTReciprocalFunction.apply(input)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.exp, torch.Tensor.exp,
                     cast=("input",))
def dt_exp(input, *, out=None):
    result = DTExpFunction.apply(input)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.log, torch.Tensor.log,
                     cast=("input",))
def dt_log(input, *, out=None):
    result = DTLogFunction.apply(input)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.prod, torch.Tensor.prod,
                     cast=("input",))
def dt_prod(input, dim=None, keepdim=False, *, out=None):
    result = DTProdFunction.apply(input, dim, keepdim)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.mean, torch.Tensor.mean,
                     cast=("input",))
def dt_mean(input, dim=None, keepdim=False, *, out=None):
    result = DTMeanFunction.apply(input, dim, keepdim)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.var, torch.Tensor.var,
                     cast=("input", "correction"))
def dt_var(input, dim=None, *, correction=1, keepdim=False, out=None):
    result = DTVarFunction.apply(input, correction, dim, keepdim)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.matmul, torch.Tensor.matmul,
                     cast=("input", "other"))
def dt_matmul(input, other, *, out=None):
    result = DTMatmulFunction.apply(input, other)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.transpose, torch.Tensor.transpose,
                     cast=("input",))
def dt_transpose(input, dim0, dim1, *, out=None):
    result = DTTransposeFunction.apply(input, dim0, dim1)

    if out is not None:
        return out.copy_(result)
    return result



import torch
from torchdt.optim import DTOptimizer

class SGD(DTOptimizer):

    def __init__(
            self,
            dtype,
            device,
            params,
            lr=0.001,
            momentum=0.0,
            dampening=0.0,
            weight_decay=0.0,
            nesterov=False,
            *,
            maximize=False,
    ):
        defaults = dict(
            lr=lr,
            momentum=momentum,
            dampening=dampening,
            weight_decay=weight_decay,
            nesterov=nesterov,
            maximize=maximize,
        )
        super().__init__(dtype, device, params, defaults)
        self.convert_params("lr", "momentum", "dampening", "weight_decay")

        self.validate_param("lr", lambda lr: lr >= torch.tensor(0.0, device=device))
        self.validate_param("momentum", lambda momentum: momentum >= torch.tensor(0.0, device=device))
        self.validate_param("dampening", lambda dampening: dampening >= torch.tensor(0.0, device=device))
        self.validate_param("weight_decay", lambda weight_decay: weight_decay >= torch.tensor(0.0, device=device))

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step."""

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group["lr"]
            momentum = group["momentum"]
            dampening = group["dampening"]
            weight_decay = group["weight_decay"]
            nesterov = group["nesterov"]
            maximize = group["maximize"]

            for p in group["params"]:
                if p.grad is None:
                    continue

                grad = p.grad
                state = self.state[p]

                if maximize:
                    grad = -grad

                if weight_decay != torch.tensor(0.0, device=self.device):
                    grad = grad + p * weight_decay

                if momentum != torch.tensor(0.0, device=self.device):
                    buf = state.get("momentum_buffer")
                    if buf is None:
                        buf = state["momentum_buffer"] = grad.clone()
                    else:
                        buf = (buf * momentum) + (grad * (torch.tensor(1.0, device=self.device) - dampening))

                    if nesterov:
                        grad = grad + buf * momentum
                    else:
                        grad = buf

                p.data.copy_(p - grad * lr)

        return loss

class TritonSGD(DTOptimizer):

    def __init__(
            self,
            dtype,
            device,
            params,
            lr=0.001,
            momentum=0.0,
            dampening=0.0,
            weight_decay=0.0,
            nesterov=False,
            *,
            maximize=False,
    ):
        defaults = dict(
            lr=lr,
            momentum=momentum,
            dampening=dampening,
            weight_decay=weight_decay,
            nesterov=nesterov,
            maximize=maximize,
        )
        super().__init__(dtype, device, params, defaults)
        self.convert_params("lr", "momentum", "dampening", "weight_decay")

        self.validate_param("lr", lambda lr: lr >= torch.tensor(0.0, device=device))
        self.validate_param("momentum", lambda momentum: momentum >= torch.tensor(0.0, device=device))
        self.validate_param("dampening", lambda dampening: dampening >= torch.tensor(0.0, device=device))
        self.validate_param("weight_decay", lambda weight_decay: weight_decay >= torch.tensor(0.0, device=device))

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step."""

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group["lr"]
            momentum = group["momentum"]
            dampening = group["dampening"]
            weight_decay = group["weight_decay"]
            nesterov = group["nesterov"]
            maximize = group["maximize"]

            for p in group["params"]:
                if p.grad is None:
                    continue

                state = self.state[p]
                buf = state.get("momentum_buffer", None)

                new_buf = self.dtype.ops.triton_sgd_step(
                    p._int, p.grad._int, buf,
                    lr._int, momentum._int,
                    dampening._int, weight_decay._int,
                    nesterov, maximize
                )
                state["momentum_buffer"] = new_buf

        return loss


import torch
from torch import Tensor
from torchdt import DType

ZERO = torch.tensor(-32768, dtype=torch.int16) # smallest positive value in LNS
POS_INF = torch.tensor(32766, dtype=torch.int16) # largest positive value in LNS
NEG_INF = torch.tensor(32767, dtype=torch.int16) # largest negative value in LNS
base = 2.0 ** (2.0 ** torch.tensor(-10, dtype=torch.float64))

class LNS16(DType, bitwidth=16, cpp_backend="lns"):
    pass

@LNS16.register_op("from_float")
def lns16_from_float(ops, t: Tensor) -> Tensor:
    t = t.to(dtype=torch.float64)
    abs_t = torch.abs(t)

    log_t = torch.log(abs_t) / torch.log(base)
    # clamp to first 15 bits then bitshift to 16 bits
    packed = torch.round(log_t).to(torch.int16).clamp(-16384, 16383) << 1 | (t < 0)

    lns_t = torch.where(
        abs_t == 0, ZERO,
        torch.where(
            torch.isposinf(t), POS_INF,
            torch.where(
                torch.isneginf(t), NEG_INF,
                packed.to(torch.int16))))
    return lns_t

@LNS16.register_op("to_float")
def lns16_to_float(ops, t: Tensor) -> Tensor:
    packed = t.view(torch.int16)
    log_t = (packed >> 1)
    sign_t = torch.where((packed & 1) == 1, -1.0, 1.0).to(torch.float64)

    abs_t = torch.pow(base, log_t)
    float_t = sign_t * abs_t

    float_t = torch.where(
        packed == ZERO, 0.0,
        torch.where(
            packed == POS_INF, float('inf'),
            torch.where(
                packed == NEG_INF, float('-inf'),
                float_t)))
    return float_t.to(torch.float64)

@LNS16.register_op("add")
def lns16_add(ops, x, y):
    max_operand = torch.max(x, y)

    abs_diff = torch.abs((x >> 1) - (y >> 1))
    sign_diff = (x ^ y) & 1

    power_term = torch.pow(base, -abs_diff)
    magnitude = torch.abs(1.0 - 2.0 * sign_diff + power_term)

    log_term = torch.log(magnitude) / torch.log(base)
    sbdb = torch.round(log_term).to(torch.int16) << 1

    return torch.where(
        x == ZERO,
        y, torch.where(
            y == ZERO,
            x, torch.where(
                x == ops.neg(y),
                ZERO, max_operand + sbdb)))

@LNS16.register_op("sub")
def lns16_sub(ops, x, y):
    return ops.add(x, ops.neg(y))

@LNS16.register_op("mul")
def lns16_mul(ops, x, y):
    return torch.where(
        x == ZERO,
        ZERO, torch.where(
            y == ZERO,
            ZERO, (x + y - (y & 1)) ^ (y & 1)))

@LNS16.register_op("div")
def lns16_div(ops, x, y):
    return torch.where(
        x == ZERO,
        ZERO, torch.where(
            y == ZERO,
            torch.where(
                ops.gt(x, ZERO), POS_INF, NEG_INF),
                (x - y + (y & 1)) ^ (y & 1)))

@LNS16.register_op("pow")
def lns16_pow(ops, x, y):
    y_float = ops.to_float(y)
    return ((x & (-2)) * y_float).to(torch.int16) & (-2)

@LNS16.register_op("neg")
def lns16_neg(ops, x):
    return torch.where(x == ops.scalar_from_float(0.0), x, x ^ 1)

@LNS16.register_op("abs")
def lns16_abs(ops, x):
    return torch.where(x == ops.scalar_from_float(0.0), x, x & (-2)) # -2 is ~1

@LNS16.register_op("sign")
def lns16_sign(ops, x):
    return torch.where(
        x == ZERO, ZERO,
        torch.where(
            (x & 1) == 1,
            ops.scalar_from_float(-1.0),
            ops.scalar_from_float(1.0)))

@LNS16.register_op("ge")
def lns16_ge(ops, x, y):
    x_log, y_log = x >> 1, y >> 1
    x_sign, y_sign = x & 1, y & 1

    both_pos = (x_sign == 0) & (y_sign == 0)
    result_both_pos = torch.ge(x_log, y_log)

    x_pos_y_neg = (x_sign == 0) & (y_sign == 1)
    x_neg_y_pos = (x_sign == 1) & (y_sign == 0)

    # no need to check explicitly for both negative case, as it's the final case
    result_both_neg = torch.ge(y_log, x_log)

    return torch.where(both_pos, result_both_pos,
        torch.where(x_pos_y_neg, True,
        torch.where(x_neg_y_pos, False, result_both_neg)))

@LNS16.register_op("gt")
def lns16_gt(ops, x, y):
    x_log, y_log = x >> 1, y >> 1
    x_sign, y_sign = x & 1, y & 1

    both_pos = (x_sign == 0) & (y_sign == 0)
    result_both_pos = torch.gt(x_log, y_log)

    x_pos_y_neg = (x_sign == 0) & (y_sign == 1)
    x_neg_y_pos = (x_sign == 1) & (y_sign == 0)

    # no need to check explicitly for both negative case, as it's the final case
    result_both_neg = torch.gt(y_log, x_log)

    return torch.where(both_pos, result_both_pos,
        torch.where(x_pos_y_neg, True,
        torch.where(x_neg_y_pos, False, result_both_neg)))

@LNS16.register_op("le")
def lns16_le(ops, x, y):
    x_log, y_log = x >> 1, y >> 1
    x_sign, y_sign = x & 1, y & 1

    both_pos = (x_sign == 0) & (y_sign == 0)
    result_both_pos = torch.le(x_log, y_log)

    x_pos_y_neg = (x_sign == 0) & (y_sign == 1)
    x_neg_y_pos = (x_sign == 1) & (y_sign == 0)

    # no need to check explicitly for both negative case, as it's the final case
    result_both_neg = torch.le(y_log, x_log)

    return torch.where(both_pos, result_both_pos,
        torch.where(x_pos_y_neg, False,
        torch.where(x_neg_y_pos, True, result_both_neg)))

@LNS16.register_op("lt")
def lns16_lt(ops, x, y):
    x_log, y_log = x >> 1, y >> 1
    x_sign, y_sign = x & 1, y & 1

    both_pos = (x_sign == 0) & (y_sign == 0)
    result_both_pos = torch.lt(x_log, y_log)

    x_pos_y_neg = (x_sign == 0) & (y_sign == 1)
    x_neg_y_pos = (x_sign == 1) & (y_sign == 0)

    # no need to check explicitly for both negative case, as it's the final case
    result_both_neg = torch.lt(y_log, x_log)

    return torch.where(both_pos, result_both_pos,
        torch.where(x_pos_y_neg, False,
        torch.where(x_neg_y_pos, True, result_both_neg)))


import torch
from torchdt.autograd import DTFunction
from torchdt.ops import register_base_op

@register_base_op("broadcast_to")
def dt_broadcast_to(ops, x, size):
    return torch.broadcast_to(x, size)

class DTBroadcastToFunction(DTFunction):

    @staticmethod
    def forward(ops, x, shape):
        return ops.broadcast_to(x, shape)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, shape = inputs
        ctx.save_for_backward(x)
        ctx.shape = shape

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, = ctx.saved_tensors

        # Sum over the broadcasted dimensions
        # First, handle prepended dimensions (when original tensor had fewer dims)
        ndims_added = grad_output.ndim - len(x.shape)
        grad_x = grad_output
        for i in range(ndims_added):
            grad_x = ops.sum(grad_x, dim=0, keepdim=False)

        # Then, handle expanded dimensions (where original dim was 1)
        for i, (orig_size, grad_size) in enumerate(zip(x.shape, grad_x.shape)):
            if orig_size == 1 and grad_size > 1:
                grad_x = ops.sum(grad_x, dim=i, keepdim=True)

        return grad_x, None

@register_base_op("clone")
def dt_clone(ops, x, memory_format=torch.preserve_format):
    return x.clone()

class DTCloneFunction(DTFunction):

    @staticmethod
    def forward(ops, x, memory_format=torch.preserve_format):
        return ops.clone(x, memory_format)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        pass

    @staticmethod
    def backward(ctx, ops, grad_output):
        return grad_output, None

@register_base_op("squeeze")
def dt_squeeze(ops, x, dim=None):
    return torch.squeeze(x, dim=dim)

class DTSqueezeFunction(DTFunction):

    @staticmethod
    def forward(ops, x, dim=None):
        return ops.squeeze(x, dim)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, dim = inputs
        ctx.save_for_backward(x)
        ctx.dim = dim

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, = ctx.saved_tensors
        grad_x = grad_output.view(x.shape)
        return grad_x, None

@register_base_op("unsqueeze")
def dt_unsqueeze(ops, x, dim):
    return torch.unsqueeze(x, dim=dim)

class DTUnsqueezeFunction(DTFunction):

    @staticmethod
    def forward(ops, x, dim):
        return ops.unsqueeze(x, dim)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, dim = inputs
        ctx.save_for_backward(x)
        ctx.dim = dim

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, = ctx.saved_tensors
        grad_x = grad_output.squeeze(dim=ctx.dim)
        return grad_x, None

@register_base_op("stack")
def dt_stack(ops, tensors, dim=0):
    return torch.stack(tensors, dim=dim)

class DTStackFunction(DTFunction):

    @staticmethod
    def forward(ops, dim, *tensors):
        return ops.stack(tensors, dim)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        dim, *tensors = inputs
        ctx.save_for_backward(*tensors)
        ctx.dim = dim

    @staticmethod
    def backward(ctx, ops, grad_output):
        return None, *grad_output.unbind(ctx.dim)

@register_base_op("cat")
def dt_cat(ops, tensors, dim=0):
    return torch.cat(tensors, dim=dim).clone()

class DTCatFunction(DTFunction):

    @staticmethod
    def forward(ops, dim, *tensors):
        return ops.cat(tensors, dim)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        dim, *tensors = inputs
        ctx.sizes = [t.size(dim) for t in tensors]
        ctx.dim = dim

    @staticmethod
    def backward(ctx, ops, grad_output):
        grad_tensors = torch.split(grad_output, ctx.sizes, dim=ctx.dim)
        return None, *grad_tensors

@register_base_op("chunk")
def dt_chunk(ops, x, chunks, dim=0):
    return tuple(y.clone() for y in torch.chunk(x, chunks, dim=dim))

class DTChunkFunction(DTFunction):

    @staticmethod
    def forward(ops, x, chunks, dim=0):
        return ops.chunk(x, chunks, dim)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        _, _, dim = inputs
        ctx.dim = dim
        ctx.out_shapes = [o.shape for o in output]
        ctx.set_materialize_grads(False)

    @staticmethod
    def backward(ctx, ops, *grad_outputs):
        parts = []
        for g, shape in zip(grad_outputs, ctx.out_shapes):
            if g is None:
                g = ops.zeros(*shape)
            parts.append(g)

        return torch.cat(parts, dim=ctx.dim), None, None

@register_base_op("where")
def dt_where(ops, condition, x, y):
    return torch.where(condition, x, y)

class DTWhereFunction(DTFunction):

    @staticmethod
    def forward(ops, condition, x, y):
        return ops.where(condition, x, y)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        condition, x, y = inputs
        ctx.save_for_backward(condition, x, y)

    @staticmethod
    def backward(ctx, ops, grad_output):
        condition, x, y = ctx.saved_tensors
        grad_x = ops.sum_to_size(torch.where(condition, grad_output, ops.scalar_from_float(0.0)), x.shape)
        grad_y = ops.sum_to_size(torch.where(condition, ops.scalar_from_float(0.0), grad_output), y.shape)
        return None, grad_x, grad_y

def _unpad_along_dim(ops, g, left, right, dim, mode):
    if left == right == 0:
        return g
    if mode == "constant":
        return g.narrow(dim, left, g.size(dim) - left - right).clone()

    interior_len = g.size(dim) - left - right
    grad_x = g.narrow(dim, left, interior_len).clone()
    first = 0
    last = interior_len - 1

    if mode == "replicate":
        if left:
            grad_x.select(dim, first).copy_(ops.add(
                grad_x.select(dim, first),
                ops.sum(g.narrow(dim, 0, left), dim=dim),
            ))
        if right:
            grad_x.select(dim, last).copy_(ops.add(
                grad_x.select(dim, last),
                ops.sum(g.narrow(dim, g.size(dim) - right, right), dim=dim),
            ))

    elif mode == "reflect":
        for i in range(left):
            target = left - i
            grad_x.select(dim, target).copy_(ops.add(
                grad_x.select(dim, target),
                g.select(dim, i),
            ))
        for i in range(right):
            target = last - 1 - i
            grad_x.select(dim, target).copy_(ops.add(
                grad_x.select(dim, target),
                g.select(dim, g.size(dim) - 1 - i),
            ))

    elif mode == "circular":
        if left:
            grad_x.narrow(dim, interior_len-left, left).copy_(ops.add(
                grad_x.narrow(dim, interior_len - left, left),
                g.narrow(dim, 0, left),
            ))
        if right:
            grad_x.narrow(dim, 0, right).copy_(ops.add(
                grad_x.narrow(dim, 0, right),
                g.narrow(dim, g.size(dim) - right, right),
            ))

    return grad_x

@register_base_op("pad")
def dt_pad(ops, input, pad, mode="constant", value=None):
    return torch.nn.functional.pad(input, pad, mode=mode, value=value)

class DTPadFunction(DTFunction):

    @staticmethod
    def forward(ops, input, pad, mode="constant", value=None):
        return ops.pad(input, pad, mode=mode, value=value)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        _, pad, mode, _ = inputs
        ctx.pad = pad
        ctx.mode = mode

    @staticmethod
    def backward(ctx, ops, grad_output):
        ndim_pad = len(ctx.pad) // 2
        grad_x = grad_output

        for i in range(ndim_pad):
            left = ctx.pad[2 * i]
            right = ctx.pad[2 * i + 1]
            dim = grad_output.dim() - 1 - i
            grad_x = _unpad_along_dim(ops, grad_x, left, right, dim, ctx.mode)

        return grad_x, None, None, None

@register_base_op("getitem")
def dt_getitem(ops, x, index):
    return x[index]

class DTGetItemFunction(DTFunction):

    @staticmethod
    def forward(ops, x, index):
        return ops.getitem(x, index)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, idx = inputs
        ctx.is_idx_tensor = torch.is_tensor(idx)
        if ctx.is_idx_tensor:
            ctx.save_for_backward(x, idx)
        else:
            ctx.save_for_backward(x)
            ctx.idx = idx

    @staticmethod
    def backward(ctx, ops, grad_output):
        if ctx.is_idx_tensor:
            x, idx = ctx.saved_tensors
        else:
            x, = ctx.saved_tensors
            idx = ctx.idx

        grad_x = torch.full_like(x, ops.scalar_from_float(0.0))
        grad_x[idx] = grad_output
        return grad_x, None

@register_base_op("setitem")
def dt_setitem(ops, x, index, value):
    return torch.index_put(x, index, value)

class DTSetItemFunction(DTFunction):

    @staticmethod
    def forward(ops, x, index, value):
        return ops.setitem(x, index, value)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        _, idx, value, _ = inputs
        ctx.is_idx_tensor = torch.is_tensor(idx)
        if ctx.is_idx_tensor:
            ctx.save_for_backward(idx, value)
        else:
            ctx.save_for_backward(value)
            ctx.idx = idx

    @staticmethod
    def backward(ctx, ops, grad_output):
        if ctx.is_idx_tensor:
            idx, value = ctx.saved_tensors
        else:
            value, = ctx.saved_tensors
            idx = ctx.idx

        grad_x = grad_output.clone()
        grad_x[idx] = ops.scalar_from_float(0.0)
        grad_value = grad_output.clone()[idx]

        if grad_value.shape != value.shape:
            # Find the dims that were broadcast (= size 1 in value but >1 in grad_value)
            extra_dims = (
                [i for i, (gv, v) in enumerate(zip(grad_value.shape[-len(value.shape):],
                                                   value.shape)) if v == 1 and gv != 1]
                + list(range(len(grad_value.shape) - len(value.shape)))  # leading dims
            )
            grad_value = ops.sum(grad_value, dim=extra_dims, keepdim=True)
            grad_value = grad_value.reshape(value.shape)

        return grad_x, None, grad_value, None

@register_base_op("to")
def dt_to(ops, x, device):
    return x.to(device=device)

class DTToFunction(DTFunction):

    @staticmethod
    def forward(ops, x, device):
        return ops.to(x, device)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, _ = inputs
        ctx.orig_device = x.device

    @staticmethod
    def backward(ctx, ops, grad_output):
        return grad_output.to(ctx.orig_device), None

@register_base_op("view")
def dt_view(ops, x, shape):
    return x.view(shape)

class DTViewFunction(DTFunction):

    @staticmethod
    def forward(ops, x, shape):
        return ops.view(x, shape)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, shape = inputs
        ctx.original_shape = x.shape
        ctx.n_shape = len(shape)

    @staticmethod
    def backward(ctx, ops, grad_output):
        grad_x = grad_output.contiguous().view(ctx.original_shape)
        return grad_x, None

@register_base_op("contiguous")
def dt_contiguous(ops, x, memory_format):
    return x.contiguous(memory_format=memory_format)

class DTContiguousFunction(DTFunction):

    @staticmethod
    def forward(ops, x, memory_format=torch.preserve_format):
        return ops.contiguous(x, memory_format)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, _ = inputs
        ctx.original_shape = x.shape
        ctx.original_strides = x.stride()

    @staticmethod
    def backward(ctx, ops, grad_output):
        return torch.as_strided(grad_output, ctx.original_shape, ctx.original_strides), None

@register_base_op("repeat")
def dt_repeat(ops, x, repeats):
    return x.repeat(*repeats)

class DTRepeatFunction(DTFunction):

    @staticmethod
    def forward(ops, x, repeats):
        return ops.repeat(x, repeats)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, repeats = inputs
        ctx.input_shape = tuple(x.shape)
        ctx.repeats = tuple(repeats)

    @staticmethod
    def backward(ctx, ops, grad_output):
        grad_x = grad_output
        for dim, rep in enumerate(ctx.repeats):
            if rep == 1:
                continue

            new_shape = list(grad_x.shape)
            new_shape[dim] = ctx.input_shape[dim]
            new_shape.insert(dim + 1, rep)
            grad_x = ops.sum(grad_x.view(*new_shape), dim=dim+1)

        return grad_x, None

@register_base_op("flatten")
def dt_flatten(ops, x, start_dim=0, end_dim=-1):
    return torch.flatten(x, start_dim=start_dim, end_dim=end_dim).clone()

class DTFlattenFunction(DTFunction):

    @staticmethod
    def forward(ops, x, start_dim=0, end_dim=-1):
        return ops.flatten(x, start_dim, end_dim)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, _, _ = inputs
        ctx.original_shape = x.shape

    @staticmethod
    def backward(ctx, ops, grad_output):
        grad_x = grad_output.reshape(ctx.original_shape)
        return grad_x, None, None

@register_base_op("reshape")
def dt_reshape(ops, x, shape):
    return torch.reshape(x, shape).clone()

class DTReshapeFunction(DTFunction):

    @staticmethod
    def forward(ops, x, shape):
        return ops.reshape(x, shape)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, _ = inputs
        ctx.original_shape = x.shape

    @staticmethod
    def backward(ctx, ops, grad_output):
        grad_x = grad_output.reshape(ctx.original_shape)
        return grad_x, None


import torch
from torchdt import DType
from torchdt.ops.unary_ops import (
    DTSignFunction,
    DTNegFunction,
    DTAbsFunction,
)

@DType.register_func(torch.sign, torch.Tensor.sign,
                     cast=("input",))
def dt_sign(input, *, out=None):
    result = DTSignFunction.apply(input)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.neg, torch.Tensor.neg,
                     cast=("input",))
def dt_neg(input, *, out=None):
    result = DTNegFunction.apply(input)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.abs, torch.Tensor.abs,
                     cast=("input",))
def dt_abs(input, *, out=None):
    result = DTAbsFunction.apply(input)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.positive, torch.Tensor.positive,
                     cast=("input",))
def dt_positive(input, *, out=None):
    if out is not None:
        return out.copy_(input)
    return input


import torch
from torchdt.optim import DTOptimizer

class Adam(DTOptimizer):

    def __init__(
            self,
            dtype,
            device,
            params,
            lr=0.001,
            betas=(0.9, 0.999),
            eps=1e-8,
            weight_decay=0.0,
            amsgrad=False,
            *,
            maximize=False,
    ):
        defaults = dict(
            lr=lr,
            beta1=betas[0],
            beta2=betas[1],
            eps=eps,
            weight_decay=weight_decay,
            amsgrad=amsgrad,
            maximize=maximize,
        )
        super().__init__(dtype, device, params, defaults)
        self.convert_params("lr", "beta1", "beta2", "eps", "weight_decay")

        self.validate_param("lr", lambda lr: lr >= torch.tensor(0.0, device=device))
        self.validate_param("eps", lambda eps: eps > torch.tensor(0.0, device=device))
        self.validate_param("beta1",lambda beta1: torch.tensor(0.0, device=device) <= beta1 < torch.tensor(1.0, device=device))
        self.validate_param("beta2",lambda beta2: torch.tensor(0.0, device=device) <= beta2 < torch.tensor(1.0, device=device))
        self.validate_param("weight_decay", lambda weight_decay: weight_decay >= torch.tensor(0.0, device=device))

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step."""

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group["lr"]
            beta1 = group["beta1"]
            beta2 = group["beta2"]
            eps = group["eps"]
            weight_decay = group["weight_decay"]
            amsgrad = group["amsgrad"]
            maximize = group["maximize"]

            for p in group["params"]:
                if p.grad is None:
                    continue

                grad = p.grad
                state = self.state[p]

                if maximize:
                    grad = -grad

                if weight_decay != torch.tensor(0.0, device=self.device):
                    grad = grad + p * weight_decay

                if len(state) == 0:
                    # First time we see this parameter
                    state["step"] = self.dtype(0, device=self.device)
                    state["exp_avg"] = torch.zeros_like(p, dtype=self.dtype, device=self.device)
                    state["exp_avg_sq"] = torch.zeros_like(p, dtype=self.dtype, device=self.device)
                    if amsgrad:
                        state["max_exp_avg_sq"] = torch.zeros_like(p, dtype=self.dtype, device=self.device)

                exp_avg = state["exp_avg"]
                exp_avg_sq = state["exp_avg_sq"]
                step = state["step"] + torch.tensor(1.0, device=self.device)

                exp_avg = (exp_avg * beta1) + (grad * (torch.tensor(1.0, device=self.device) - beta1))
                exp_avg_sq = (exp_avg_sq * beta2) + (grad * grad * (torch.tensor(1.0, device=self.device) - beta2))

                bias_corr1 = torch.tensor(1.0, device=self.device) - beta1 ** step
                bias_corr2 = torch.tensor(1.0, device=self.device) - beta2 ** step

                if amsgrad:
                    max_exp_avg_sq = torch.maximum(state["max_exp_avg_sq"], exp_avg_sq)
                    state["max_exp_avg_sq"] = max_exp_avg_sq
                    v_denom = max_exp_avg_sq
                else:
                    v_denom = exp_avg_sq

                step_size = lr * torch.sqrt(bias_corr2) / bias_corr1
                p.data.copy_(p - step_size * exp_avg / (torch.sqrt(v_denom) + eps))

                state["exp_avg"] = exp_avg
                state["exp_avg_sq"] = exp_avg_sq

        return loss


class TritonAdam(DTOptimizer):

    def __init__(
            self,
            dtype,
            device,
            params,
            lr=0.001,
            betas=(0.9, 0.999),
            eps=1e-8,
            weight_decay=0.0,
            amsgrad=False,
            *,
            maximize=False,
    ):
        defaults = dict(
            lr=lr,
            beta1=betas[0],
            beta2=betas[1],
            eps=eps,
            weight_decay=weight_decay,
            amsgrad=amsgrad,
            maximize=maximize,
        )
        super().__init__(dtype, device, params, defaults)
        self.convert_params("lr", "beta1", "beta2", "eps", "weight_decay")

        self.validate_param("lr", lambda lr: lr >= torch.tensor(0.0, device=device))
        self.validate_param("eps", lambda eps: eps > torch.tensor(0.0, device=device))
        self.validate_param("beta1",lambda beta1: torch.tensor(0.0, device=device) <= beta1 < torch.tensor(1.0, device=device))
        self.validate_param("beta2",lambda beta2: torch.tensor(0.0, device=device) <= beta2 < torch.tensor(1.0, device=device))
        self.validate_param("weight_decay", lambda weight_decay: weight_decay >= torch.tensor(0.0, device=device))

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step."""

        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group["lr"]
            beta1 = group["beta1"]
            beta2 = group["beta2"]
            eps = group["eps"]
            weight_decay = group["weight_decay"]
            amsgrad = group["amsgrad"]
            maximize = group["maximize"]

            for p in group["params"]:
                if p.grad is None:
                    continue

                grad = p.grad
                state = self.state[p]

                if len(state) == 0:
                    # First time we see this parameter
                    state["step"] = self.dtype(0, device=self.device)
                    state["exp_avg"] = torch.zeros_like(p, dtype=self.dtype, device=self.device)._int
                    state["exp_avg_sq"] = torch.zeros_like(p, dtype=self.dtype, device=self.device)._int
                    if amsgrad:
                        state["max_exp_avg_sq"] = torch.zeros_like(p, dtype=self.dtype, device=self.device)._int

                step = state["step"] + torch.tensor(1.0, device=self.device)
                exp_avg = state["exp_avg"]
                exp_avg_sq = state["exp_avg_sq"]
                max_exp_avg_sq = state.get("max_exp_avg_sq", None)

                bias_corr1 = torch.tensor(1.0, device=self.device) - (beta1 ** step)
                bias_corr2 = torch.tensor(1.0, device=self.device) - (beta2 ** step)

                exp_avg, exp_avg_sq, max_exp_avg_sq = self.dtype.ops.triton_adam_step(
                    p._int, grad._int,
                    exp_avg, exp_avg_sq, max_exp_avg_sq,
                    lr._int, beta1._int, beta2._int,
                    eps._int, weight_decay._int,
                    bias_corr1._int, bias_corr2._int,
                    amsgrad, maximize,
                )

                state["exp_avg"] = exp_avg
                state["exp_avg_sq"] = exp_avg_sq
                if amsgrad:
                    state["max_exp_avg_sq"] = max_exp_avg_sq

        return loss


import torch
from torchdt import DType
from torchdt.ops.activation_ops import (
    DTReLUFunction,
    DTLeakyReLUFunction,
    DTThresholdFunction,
    DTTanhFunction,
    DTSigmoidFunction,
    DTLogSigmoidFunction,
    DTSoftminFunction,
    DTSoftmaxFunction,
    DTLogSoftmaxFunction,
    DTHardtanhFunction,
    DTGluFunction,
)

@DType.register_func(torch.nn.functional.relu, torch.Tensor.relu,
                     cast=("input",))
def dt_relu(input, inplace=False):
    result = DTReLUFunction.apply(input)
    return result

@DType.register_func(torch.nn.functional.leaky_relu,
                     cast=("input", "negative_slope"))
def dt_leaky_relu(input, negative_slope=0.01, inplace=False):
    result = DTLeakyReLUFunction.apply(input, negative_slope)
    return result

@DType.register_func(torch.nn.functional.threshold,
                     cast=("input", "threshold", "value"))
def dt_threshold(input, threshold, value, inplace=False):
    result = DTThresholdFunction.apply(input, threshold, value)
    return result

@DType.register_func(torch.nn.functional.tanh, torch.Tensor.tanh,
                     cast=("input",))
def dt_tanh(input, *, out=None):
    result = DTTanhFunction.apply(input)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.nn.functional.sigmoid, torch.Tensor.sigmoid,
                     cast=("input",))
def dt_sigmoid(input, *, out=None):
    result = DTSigmoidFunction.apply(input)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.nn.functional.logsigmoid,
                     cast=("input",))
def dt_logsigmoid(input, *, out=None):
    result = DTLogSigmoidFunction.apply(input)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.nn.functional.softmin,
                     cast=("input",))
def dt_softmin(input, dim=None, *, out=None):
    result = DTSoftminFunction.apply(input, dim)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.nn.functional.softmax, torch.Tensor.softmax,
                     cast=("input",))
def dt_softmax(input, dim=None, *, out=None):
    result = DTSoftmaxFunction.apply(input, dim)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.nn.functional.log_softmax, torch.Tensor.log_softmax,
                     cast=("input",))
def dt_log_softmax(input, dim=None, _stacklevel=3, dtype=None, *, out=None):
    result = DTLogSoftmaxFunction.apply(input, dim)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.nn.functional.hardtanh,
                     cast=("input", "min_val", "max_val"))
def dt_hardtanh(input, min_val=-1.0, max_val=1.0, *, out=None):
    result = DTHardtanhFunction.apply(input, min_val, max_val)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.nn.functional.glu,
                     cast=("input",))
def dt_glu(input, dim=-1, *, out=None):
    result = DTGluFunction.apply(input, dim)

    if out is not None:
        return out.copy_(result)
    return result


import math
import torch
from torchdt.autograd import DTFunction
from torchdt.ops import register_base_op

@register_base_op("mse_loss")
def dt_mse_loss(ops, x, y, reduction='mean', weight=None):
    errors = ops.sub(x, y)
    squared_errors = ops.square(errors)

    if weight is not None:
        squared_errors = ops.mul(squared_errors, weight)

    if reduction == 'none':
        return squared_errors

    elif reduction == 'sum':
        squared_error_sum = ops.sum(squared_errors)
        return squared_error_sum

    elif reduction == 'mean':
        squared_error_sum = ops.sum(squared_errors)

        if weight is not None:
            weight_sum = ops.sum(weight)
            weighted_mean = ops.div(squared_error_sum, weight_sum)
            return weighted_mean

        else:
            num_elements = x.numel()
            mean = ops.div(squared_error_sum, ops.scalar_from_float(num_elements))
            return mean

class DTMSELossFunction(DTFunction):

    @staticmethod
    def forward(ops, x, y, reduction='mean', weight=None):
        return ops.mse_loss(x, y, reduction, weight)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, y, reduction, weight = inputs
        ctx.save_for_backward(x, y, weight)
        ctx.reduction = reduction

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, y, weight = ctx.saved_tensors

        if weight is not None:
            grad = ops.sub(x, y)
            grad = ops.mul(grad, ops.scalar_from_float(2.0))
            grad = ops.mul(grad, weight)

            if ctx.reduction == 'mean':
                weight_sum = ops.sum(weight)
                grad = ops.div(grad, weight_sum)

        else:
            grad = ops.sub(x, y)
            grad = ops.mul(grad, ops.scalar_from_float(2.0))

            if ctx.reduction == 'mean':
                num_elements = x.numel()
                grad = ops.div(grad, ops.scalar_from_float(num_elements))

        grad_x = ops.mul(grad, grad_output)
        grad_y = ops.neg(grad_x)
        return grad_x, grad_y, None, None

@register_base_op("l1_loss")
def dt_l1_loss(ops, x, y, reduction='mean', weight=None):
    errors = ops.sub(x, y)
    abs_errors = ops.abs(errors)

    if weight is not None:
        abs_errors = ops.mul(abs_errors, weight)

    if reduction == 'none':
        return abs_errors

    elif reduction == 'sum':
        abs_error_sum = ops.sum(abs_errors)
        return abs_error_sum

    elif reduction == 'mean':
        abs_error_sum = ops.sum(abs_errors)

        if weight is not None:
            weight_sum = ops.sum(weight)
            weighted_mean = ops.div(abs_error_sum, weight_sum)
            return weighted_mean

        else:
            num_elements = x.numel()
            mean = ops.div(abs_error_sum, ops.scalar_from_float(num_elements))
            return mean

class DTL1LossFunction(DTFunction):

    @staticmethod
    def forward(ops, x, y, reduction='mean', weight=None):
        return ops.l1_loss(x, y, reduction, weight)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, y, reduction, weight = inputs
        ctx.save_for_backward(x, y, weight)
        ctx.reduction = reduction

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, y, weight = ctx.saved_tensors

        sign = ops.sign(ops.sub(x, y))

        if weight is not None:
            grad = ops.mul(sign, weight)

            if ctx.reduction == 'mean':
                weight_sum = ops.sum(weight)
                grad = ops.div(grad, weight_sum)

        else:
            grad = sign

            if ctx.reduction == 'mean':
                num_elements = x.numel()
                grad = ops.div(grad, ops.scalar_from_float(num_elements))

        grad_x = ops.mul(grad, grad_output)
        grad_y = ops.neg(grad_x)
        return grad_x, grad_y, None, None

@register_base_op("binary_cross_entropy")
def dt_binary_cross_entropy(ops, x, y, weight=None, reduction='mean'):
    log_x = ops.log(x)
    pos_log_prob = ops.mul(y, log_x)
    x2 = ops.sub(ops.scalar_from_float(1.0), x)
    log_x2 = ops.log(x2)
    y2 = ops.sub(ops.scalar_from_float(1.0), y)
    neg_log_prob = ops.mul(y2, log_x2)

    loss = ops.add(pos_log_prob, neg_log_prob)
    if weight is not None:
        loss = ops.mul(loss, weight)
    loss = ops.neg(loss)

    if reduction == 'none':
        return loss

    elif reduction == 'sum':
        loss_sum = ops.sum(loss)
        return loss_sum

    elif reduction == 'mean':
        loss_sum = ops.sum(loss)

        if weight is not None:
            weight_sum = ops.sum(weight)
            weighted_mean = ops.div(loss_sum, weight_sum)
            return weighted_mean

        else:
            num_elements = x.numel()
            mean = ops.div(loss_sum, ops.scalar_from_float(num_elements))
            return mean

class DTBCELossFunction(DTFunction):

    @staticmethod
    def forward(ops, x, y, weight=None, reduction='mean'):
        return ops.binary_cross_entropy(x, y, weight, reduction)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, y, weight, reduction = inputs
        ctx.save_for_backward(x, y, weight)
        ctx.reduction = reduction

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, y, weight = ctx.saved_tensors

        if weight is not None:
            one_minus_x = ops.sub(ops.scalar_from_float(1.0), x)
            one_minus_y = ops.sub(ops.scalar_from_float(1.0), y)
            term1 = ops.div(one_minus_y, one_minus_x)
            term2 = ops.div(y, x)

            grad_x = ops.sub(term1, term2)
            grad_x = ops.mul(grad_x, weight)

            grad_y = ops.div(x, one_minus_x)
            grad_y = ops.log(grad_y)
            grad_y = ops.mul(grad_y, weight)
            grad_y = ops.neg(grad_y)

            if ctx.reduction == 'mean':
                weight_sum = ops.sum(weight)
                grad_x = ops.div(grad_x, weight_sum)
                grad_y = ops.div(grad_y, weight_sum)

        else:
            one_minus_x = ops.sub(ops.scalar_from_float(1.0), x)
            one_minus_y = ops.sub(ops.scalar_from_float(1.0), y)
            term1 = ops.div(one_minus_y, one_minus_x)
            term2 = ops.div(y, x)

            grad_x = ops.sub(term1, term2)
            grad_y = ops.div(x, one_minus_x)
            grad_y = ops.log(grad_y)
            grad_y = ops.neg(grad_y)

            if ctx.reduction == 'mean':
                num_elements = ops.scalar_from_float(x.numel())
                grad_x = ops.div(grad_x, num_elements)
                grad_y = ops.div(grad_y, num_elements)

        grad_x = ops.mul(grad_x, grad_output)
        grad_y = ops.mul(grad_y, grad_output)
        return grad_x, grad_y, None, None

@register_base_op("binary_cross_entropy_with_logits")
def dt_binary_cross_entropy_with_logits(ops, x, y, weight=None, reduction='mean', pos_weight=None):
    if pos_weight is not None:
        raise NotImplementedError("pos_weight is not currently implemented.")

    sigmoid_x = ops.sigmoid(x)
    log_sigmoid_x = ops.log(sigmoid_x)
    pos_log_prob = ops.mul(y, log_sigmoid_x)

    sigmoid_x2 = ops.sub(ops.scalar_from_float(1.0), sigmoid_x)
    log_sigmoid_x2 = ops.log(sigmoid_x2)
    y2 = ops.sub(ops.scalar_from_float(1.0), y)
    neg_log_prob = ops.mul(y2, log_sigmoid_x2)

    loss = ops.add(pos_log_prob, neg_log_prob)
    if weight is not None:
        loss = ops.mul(loss, weight)
    loss = ops.neg(loss)

    if reduction == 'none':
        return loss

    elif reduction == 'sum':
        loss_sum = ops.sum(loss)
        return loss_sum

    elif reduction == 'mean':
        loss_sum = ops.sum(loss)

        if weight is not None:
            weight_sum = ops.sum(weight)
            weighted_mean = ops.div(loss_sum, weight_sum)
            return weighted_mean

        else:
            num_elements = ops.scalar_from_float(x.numel())
            mean = ops.div(loss_sum, num_elements)
            return mean

class DTBCEWithLogitsLossFunction(DTFunction):
    
    @staticmethod
    def forward(ops, x, y, weight=None, reduction='mean', pos_weight=None):
        return ops.binary_cross_entropy_with_logits(x, y, weight, reduction, pos_weight)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, y, weight, reduction = inputs
        ctx.save_for_backward(x, y, weight)
        ctx.reduction = reduction

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, y, weight = ctx.saved_tensors

        if weight is not None:
            sigmoid_x = ops.sigmoid(x)
            grad_x = ops.sub(sigmoid_x, y)
            grad_x = ops.mul(grad_x, weight)

            grad_y = ops.mul(x, weight)
            grad_y = ops.neg(grad_y)

            if ctx.reduction == 'mean':
                weight_sum = ops.sum(weight)
                grad_x = ops.div(grad_x, weight_sum)
                grad_y = ops.div(grad_y, weight_sum)

        else:
            sigmoid_x = ops.sigmoid(x)
            grad_x = ops.sub(sigmoid_x, y)
            grad_y = ops.neg(x)

            if ctx.reduction == 'mean':
                num_elements = ops.scalar_from_float(x.numel())
                grad_x = ops.div(grad_x, num_elements)
                grad_y = ops.div(grad_y, num_elements)

        grad_x = ops.mul(grad_x, grad_output)
        grad_y = ops.mul(grad_y, grad_output)

        return grad_x, grad_y, None, None

@register_base_op("nll_loss")
def dt_nll_loss(ops, x, y, weight=None, reduction='mean', ignore_index=-100):
    if ignore_index != -100:
        raise NotImplementedError("ignore_index is not currently implemented.")

    if x.dim() == 1:
        nll = x[y]
    else:
        nll = x.gather(1, y.view(-1, 1)).squeeze(1)

    if weight is not None:
        sample_weights = weight[y]
        nll = ops.mul(nll, sample_weights)
    loss = ops.neg(nll)

    if reduction == 'none':
        return loss

    elif reduction == 'sum':
        loss_sum = ops.sum(loss)
        return loss_sum

    elif reduction == 'mean':
        loss_sum = ops.sum(loss)

        if weight is not None:
            weight_sum = ops.sum(sample_weights)
            weighted_mean = ops.div(loss_sum, weight_sum)
            return weighted_mean

        else:
            batch_size = ops.scalar_from_float(y.size(0))
            mean = ops.div(loss_sum, batch_size)
            return mean

class DTNLLLossFunction(DTFunction):

    @staticmethod
    def forward(ops, x, y, weight=None, reduction='mean', ignore_index=-100):
        return ops.nll_loss(x, y, weight, reduction, ignore_index)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, y, weight, reduction, _ = inputs
        ctx.save_for_backward(x, y, weight)
        ctx.reduction = reduction

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, y, weight = ctx.saved_tensors

        if weight is not None:
            grad_x = ops.zeros_like(x)
            if grad_x.dim() == 1:
                grad_x[y] = ops.neg(weight[y])

            else:
                batch_size = y.size(0)
                indices = torch.arange(batch_size)
                grad_x[indices, y] = ops.neg(weight[y])

            if ctx.reduction == 'mean':
                weight_sum = ops.sum(weight)
                grad_x = ops.div(grad_x, weight_sum)

        else:
            grad_x = ops.zeros_like(x)
            if grad_x.dim() == 1:
                grad_x[y] = ops.scalar_from_float(-1.0)

            else:
                batch_size = y.size(0)
                indices = torch.arange(batch_size)
                grad_x[indices, y] = ops.scalar_from_float(-1.0)

            if ctx.reduction == 'mean':
                batch_size = ops.scalar_from_float(y.size(0))
                grad_x = ops.div(grad_x, batch_size)

        if ctx.reduction == 'none':
            if grad_x.dim() == 1:
                grad_x = ops.mul(grad_x, grad_output)

            else:
                batch_size = y.size(0)
                indices = torch.arange(batch_size)
                grad_x[indices, y] = ops.mul(grad_x[indices, y], grad_output)

        else:
            grad_x = ops.mul(grad_x, grad_output)

        return grad_x, None, None, None, None

@register_base_op("poisson_nll_loss")
def dt_poisson_nll_loss(ops, x, y, eps, log_input=True, full=False, reduction='mean'):
    if log_input:
        exp_x = ops.exp(x)
        loss = ops.sub(exp_x, ops.mul(y, x))
    else:
        log_x = ops.log(ops.add(x, eps))
        loss = ops.sub(x, ops.mul(y, log_x))

    if full:
        y_clamped = torch.where(ops.gt(y, ops.scalar_from_float(1.0)), y, ops.scalar_from_float(1.0))

        two_pi = ops.scalar_from_float(2.0 * math.pi)
        stirling_term1 = ops.mul(y_clamped, ops.log(y_clamped))
        stirling_term3 = ops.mul(ops.log(ops.mul(two_pi, y_clamped)), ops.scalar_from_float(0.5))
        stirling = ops.add(ops.sub(stirling_term1, y_clamped), stirling_term3)

        loss = ops.add(loss, stirling)

    if reduction == 'none':
        return loss

    elif reduction == 'sum':
        loss_sum = ops.sum(loss)
        return loss_sum

    elif reduction == 'mean':
        loss_sum = ops.sum(loss)
        num_elements = ops.scalar_from_float(x.numel())
        mean = ops.div(loss_sum, num_elements)
        return mean

class DTPoissonNLLLossFunction(DTFunction):

    @staticmethod
    def forward(ops, x, y, eps, log_input=True, full=False, reduction='mean'):
        return ops.poisson_nll_loss(x, y, eps, log_input, full, reduction)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, y, eps, log_input, full, reduction = inputs
        ctx.save_for_backward(x, y, eps)
        ctx.log_input = log_input
        ctx.full = full
        ctx.reduction = reduction

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, y, eps = ctx.saved_tensors

        if ctx.log_input:
            grad_x = ops.sub(ops.exp(x), y)
            grad_y = ops.neg(x)

        else:
            grad_x = ops.div(y, ops.add(x, eps))
            grad_x = ops.sub(ops.scalar_from_float(1.0), grad_x)
            grad_y = ops.neg(ops.log(ops.add(x, eps)))

        if ctx.full:
            stirling_grad = torch.where(ops.gt(y, ops.scalar_from_float(1.0)),
                                        ops.add(ops.log(y), ops.div(
                                            ops.scalar_from_float(0.5), y)),
                                       ops.scalar_from_float(0.0))
            grad_y = ops.add(grad_y, stirling_grad)

        if ctx.reduction == 'mean':
            num_elements = ops.scalar_from_float(x.numel())
            grad_x = ops.div(grad_x, num_elements)

        grad_x = ops.mul(grad_x, grad_output)
        grad_y = ops.mul(grad_y, grad_output)

        return grad_x, grad_y, None, None, None, None

@register_base_op("hinge_embedding_loss")
def dt_hinge_embedding_loss(ops, x, y, margin=None, reduction='mean'):
    positive_mask = ops.eq(y, ops.scalar_from_float(1.0))
    loss = torch.where(positive_mask, x, ops.maximum(ops.scalar_from_float(0.0), ops.sub(margin, x)))

    if reduction == 'none':
        return loss

    elif reduction == 'sum':
        loss_sum = ops.sum(loss)
        return loss_sum

    elif reduction == 'mean':
        loss_sum = ops.sum(loss)
        num_elements = ops.scalar_from_float(x.numel())
        mean = ops.div(loss_sum, num_elements)
        return mean

class DTHingeEmbeddingLossFunction(DTFunction):

    @staticmethod
    def forward(ops, x, y, margin, reduction='mean'):
        return ops.hinge_embedding_loss(x, y, margin, reduction)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, y, margin, reduction = inputs
        ctx.save_for_backward(x, y, margin)
        ctx.reduction = reduction

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, y, margin = ctx.saved_tensors

        grad_x = torch.where(ops.eq(y, ops.scalar_from_float(1.0)),
                             ops.scalar_from_float(1.0),
                             torch.where(ops.gt(ops.sub(margin, x), ops.scalar_from_float(0.0)),
                                         ops.scalar_from_float(-1.0), ops.scalar_from_float(0.0)))

        if ctx.reduction == 'mean':
            num_elements = ops.scalar_from_float(x.numel())
            grad_x = ops.div(grad_x, num_elements)

        grad_x = ops.mul(grad_x, grad_output)

        return grad_x, None, None, None

@register_base_op("kl_div")
def dt_kl_div(ops, x, y, reduction='mean', log_target=False):
    if log_target:
        loss = ops.mul(ops.exp(y), ops.sub(y, x))
    else:
        loss = ops.mul(y, ops.sub(ops.log(y), x))

    if reduction == 'none':
        return loss

    elif reduction == 'sum':
        loss_sum = ops.sum(loss)
        return loss_sum

    elif reduction == 'mean':
        loss_sum = ops.sum(loss)
        num_elements = ops.scalar_from_float(x.numel())
        mean = ops.div(loss_sum, num_elements)
        return mean

    elif reduction == 'batchmean':
        loss_sum = ops.sum(loss)
        num_elements = ops.scalar_from_float(x.size(0))
        batch_mean = ops.div(loss_sum, num_elements)
        return batch_mean

class DTKLDivLossFunction(DTFunction):

    @staticmethod
    def forward(ops, x, y, reduction='mean', log_target=False):
        return ops.kl_div(x, y, reduction, log_target)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, y, reduction, log_target = inputs
        ctx.save_for_backward(x, y)
        ctx.reduction = reduction
        ctx.log_target = log_target

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, y = ctx.saved_tensors

        if ctx.log_target:
            exp_y = ops.exp(y)
            grad_x = ops.neg(exp_y)
            grad_y = ops.mul(exp_y, ops.add(ops.sub(y, x), ops.scalar_from_float(1.0)))
        else:
            grad_x = ops.neg(y)
            grad_y = ops.add(ops.sub(ops.log(y), x), ops.scalar_from_float(1.0))

        if ctx.reduction == 'mean':
            num_elements = ops.scalar_from_float(x.numel())
            grad_x = ops.div(grad_x, num_elements)
            grad_y = ops.div(grad_y, num_elements)

        elif ctx.reduction == 'batchmean':
            num_elements = ops.scalar_from_float(x.size(0))
            grad_x = ops.div(grad_x, num_elements)
            grad_y = ops.div(grad_y, num_elements)

        grad_x = ops.mul(grad_x, grad_output)
        grad_y = ops.mul(grad_y, grad_output)

        return grad_x, grad_y, None, None

@register_base_op("margin_ranking_loss")
def dt_margin_ranking_loss(ops, x1, x2, y, margin, reduction='mean'):
    loss = ops.sub(x1, x2)
    loss = ops.mul(loss, y)
    loss = ops.sub(margin, loss)
    loss = ops.maximum(ops.scalar_from_float(0.0), loss)

    if reduction == 'none':
        return loss

    elif reduction == 'sum':
        loss_sum = ops.sum(loss)
        return loss_sum

    elif reduction == 'mean':
        loss_sum = ops.sum(loss)
        num_elements = ops.scalar_from_float(x1.numel())
        mean = ops.div(loss_sum, num_elements)
        return mean

class DTMarginRankingLossFunction(DTFunction):

    @staticmethod
    def forward(ops, x1, x2, y, margin, reduction='mean'):
        return ops.margin_ranking_loss(x1, x2, y, margin, reduction)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x1, x2, y, margin, reduction = inputs
        ctx.save_for_backward(x1, x2, y, margin)
        ctx.reduction = reduction

    @staticmethod
    def backward(ctx, ops, grad_output):
        x1, x2, y, margin = ctx.saved_tensors

        loss = ops.sub(x1, x2)
        loss = ops.mul(loss, y)
        loss = ops.sub(margin, loss)
        gt_zero_mask = ops.gt(loss, ops.scalar_from_float(0.0))

        grad_x1 = torch.where(gt_zero_mask, ops.neg(y), ops.scalar_from_float(0.0))
        grad_x2 = torch.where(gt_zero_mask, y, ops.scalar_from_float(0.0))
        grad_y = torch.where(gt_zero_mask, ops.sub(x2, x1), ops.scalar_from_float(0.0))

        if ctx.reduction == 'mean':
            num_elements = ops.scalar_from_float(x1.numel())
            grad_x1 = ops.div(grad_x1, num_elements)
            grad_x2 = ops.div(grad_x2, num_elements)
            grad_y = ops.div(grad_y, num_elements)

        grad_x1 = ops.mul(grad_x1, grad_output)
        grad_x2 = ops.mul(grad_x2, grad_output)
        grad_y = ops.mul(grad_y, grad_output)

        return grad_x1, grad_x2, grad_y, None, None

@register_base_op("gaussian_nll_loss")
def dt_gaussian_nll_loss(ops, x, y, var, eps, full=False, reduction='mean'):
    var_eps = ops.maximum(var, eps)
    loss = ops.square(ops.sub(x, y))
    loss = ops.add(ops.log(var_eps), ops.div(loss, var_eps))

    if full:
        two_pi = ops.scalar_from_float(2.0 * math.pi)
        loss = ops.add(loss, ops.log(two_pi))

    loss = ops.div(loss, ops.scalar_from_float(2.0))

    if reduction == 'none':
        return loss

    elif reduction == 'sum':
        loss_sum = ops.sum(loss)
        return loss_sum

    elif reduction == 'mean':
        loss_sum = ops.sum(loss)
        num_elements = x.numel()
        mean = ops.div(loss_sum, ops.scalar_from_float(num_elements))
        return mean

class DTGaussianNLLLossFunction(DTFunction):

    @staticmethod
    def forward(ops, x, y, var, eps, full=False, reduction='mean'):
        return ops.gaussian_nll_loss(x, y, var, eps, full, reduction)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, y, var, eps, _, reduction = inputs
        ctx.save_for_backward(x, y, var, eps)
        ctx.reduction = reduction

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, y, var, eps = ctx.saved_tensors

        var_eps = ops.maximum(var, eps)
        grad_x = ops.div(ops.sub(x, y), var_eps)
        grad_y = ops.neg(grad_x)

        grad_var = ops.square(ops.div(ops.sub(x, y), var))
        grad_var = ops.sub(ops.reciprocal(var), grad_var)
        grad_var = ops.div(grad_var, ops.scalar_from_float(2.0))

        if ctx.reduction == 'mean':
            num_elements = ops.scalar_from_float(x.numel())
            grad_x = ops.div(grad_x, num_elements)
            grad_y = ops.div(grad_y, num_elements)
            grad_var = ops.div(grad_var, num_elements)

        grad_x = ops.mul(grad_x, grad_output)
        grad_y = ops.mul(grad_y, grad_output)
        grad_var = ops.mul(grad_var, grad_output)

        return grad_x, grad_y, grad_var, None, None, None

@register_base_op("huber_loss")
def dt_huber_loss(ops, x, y, delta, reduction='mean', weight=None):
    two = ops.scalar_from_float(2.0)

    abs_diff = ops.abs(ops.sub(x, y))
    l1_term = ops.sub(abs_diff, ops.div(delta, two))
    l1_term = ops.mul(l1_term, delta)

    l2_term = ops.square(ops.sub(x, y))
    l2_term = ops.div(l2_term, two)

    loss = torch.where(ops.lt(abs_diff, delta), l2_term, l1_term)
    if weight is not None:
        loss = ops.mul(loss, weight)

    if reduction == 'none':
        return loss

    elif reduction == 'sum':
        loss_sum = ops.sum(loss)
        return loss_sum

    elif reduction == 'mean':
        loss_sum = ops.sum(loss)
        num_elements = ops.scalar_from_float(x.numel())
        mean = ops.div(loss_sum, num_elements)
        return mean

class DTHuberLossFunction(DTFunction):

    @staticmethod
    def forward(ops, x, y, delta, reduction='mean', weight=None):
        return ops.huber_loss(x, y, delta, reduction, weight)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, y, delta, reduction, weight = inputs
        ctx.save_for_backward(x, y, delta, weight)
        ctx.reduction = reduction

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, y, delta, weight = ctx.saved_tensors
        two = ops.scalar_from_float(2.0)

        l2_loss_grad_x = ops.sub(x, y)
        l2_loss_grad_y = ops.neg(l2_loss_grad_x)
        l1_loss_grad_x = ops.mul(ops.sign(ops.sub(x, y)), delta)
        l1_loss_grad_y = ops.neg(l1_loss_grad_x)

        abs_diff = ops.abs(ops.sub(x, y))
        l2_mask = ops.lt(abs_diff, delta)
        grad_x = torch.where(l2_mask, l2_loss_grad_x, l1_loss_grad_x)
        grad_y = torch.where(l2_mask, l2_loss_grad_y, l1_loss_grad_y)

        grad_w = None
        if weight is not None:
            abs_diff = ops.abs(ops.sub(x, y))
            l1_term = ops.sub(abs_diff, ops.div(delta, two))
            l1_term = ops.mul(l1_term, delta)

            l2_term = ops.square(ops.sub(x, y))
            l2_term = ops.div(l2_term, two)

            grad_w = torch.where(ops.lt(abs_diff, delta), l2_term, l1_term)
            grad_x = ops.mul(grad_x, weight)
            grad_y = ops.mul(grad_y, weight)

        if ctx.reduction == 'mean':
            num_elements = ops.scalar_from_float(x.numel())
            grad_x = ops.div(grad_x, num_elements)
            grad_y = ops.div(grad_y, num_elements)
            if weight is not None:
                grad_w = ops.div(grad_w, num_elements)

        grad_x = ops.mul(grad_x, grad_output)
        grad_y = ops.mul(grad_y, grad_output)
        if weight is not None:
            grad_w = ops.mul(grad_w, grad_output)

        return grad_x, grad_y, None, None, grad_w

@register_base_op("smooth_l1_loss")
def dt_smooth_l1_loss(ops, x, y, beta, reduction='mean'):
    two = ops.scalar_from_float(2.0)

    abs_diff = ops.abs(ops.sub(x, y))
    l1_term = ops.sub(abs_diff, ops.div(beta, two))
    l2_term = ops.square(ops.sub(x, y))
    l2_term = ops.div(l2_term, ops.mul(two, beta))
    loss = torch.where(ops.lt(abs_diff, beta), l2_term, l1_term)

    if reduction == 'none':
        return loss

    elif reduction == 'sum':
        loss_sum = ops.sum(loss)
        return loss_sum

    elif reduction == 'mean':
        loss_sum = ops.sum(loss)
        num_elements = ops.scalar_from_float(x.numel())
        mean = ops.div(loss_sum, num_elements)
        return mean

class DTSmoothL1LossFunction(DTFunction):

    @staticmethod
    def forward(ops, x, y, beta, reduction='mean'):
        return ops.smooth_l1_loss(x, y, beta, reduction)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, y, beta, reduction = inputs
        ctx.save_for_backward(x, y, beta)
        ctx.reduction = reduction

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, y, beta = ctx.saved_tensors

        l2_loss_grad_x = ops.div(ops.sub(x, y), beta)
        l2_loss_grad_y = ops.neg(l2_loss_grad_x)
        l1_loss_grad_x = ops.sign(ops.sub(x, y))
        l1_loss_grad_y = ops.neg(l1_loss_grad_x)

        abs_diff = ops.abs(ops.sub(x, y))
        l2_mask = ops.lt(abs_diff, beta)
        grad_x = torch.where(l2_mask, l2_loss_grad_x, l1_loss_grad_x)
        grad_y = torch.where(l2_mask, l2_loss_grad_y, l1_loss_grad_y)

        if ctx.reduction == 'mean':
            num_elements = ops.scalar_from_float(x.numel())
            grad_x = ops.div(grad_x, num_elements)
            grad_y = ops.div(grad_y, num_elements)

        grad_x = ops.mul(grad_x, grad_output)
        grad_y = ops.mul(grad_y, grad_output)

        return grad_x, grad_y, None, None

@register_base_op("cross_entropy")
def dt_cross_entropy(ops, x, y, weight = None, ignore_index = -100, reduction = 'mean', label_smoothing = 0.0):
    if ignore_index != -100:
        raise NotImplementedError("ignore_index is not currently implemented.")
    if label_smoothing is not None and label_smoothing != ops.scalar_from_float(0.0):
        raise NotImplementedError("label_smoothing is not currently implemented.")

    dim = -1 if x.dim() > 1 else 0
    m = ops.max(x, dim=dim, keepdim=True)[0]
    x_sub_m = ops.sub(x, m)
    exp_x_sub_m = ops.exp(x_sub_m)
    sum_exp_x_sub_m = ops.sum(exp_x_sub_m, dim=dim, keepdim=True)
    log_sum_exp_x_sub_m = ops.log(sum_exp_x_sub_m)
    log_softmax = ops.sub(x_sub_m, log_sum_exp_x_sub_m)

    if log_softmax.dim() == 1:
        nll = log_softmax[y]
    else:
        nll = log_softmax.gather(1, y.view(-1, 1)).squeeze(1)

    if weight is not None:
        sample_weights = weight[y]
        nll = ops.mul(nll, sample_weights)
    loss = ops.neg(nll)

    if reduction == 'none':
        return loss

    elif reduction == 'sum':
        loss_sum = ops.sum(loss)
        return loss_sum

    elif reduction == 'mean':
        loss_sum = ops.sum(loss)

        if weight is not None:
            weight_sum = ops.sum(sample_weights)
            weighted_mean = ops.div(loss_sum, weight_sum)
            return weighted_mean

        else:
            batch_size = ops.scalar_from_float(y.size(0))
            mean = ops.div(loss_sum, batch_size)
            return mean

class DTCrossEntropyLossFunction(DTFunction):

    @staticmethod
    def forward(ops, x, y, weight=None, ignore_index=-100, reduction='mean', label_smoothing=None):
        return ops.cross_entropy(x, y, weight, ignore_index, reduction, label_smoothing)

    @staticmethod
    def setup_context(ctx, ops, inputs, output):
        x, y, weight, _, reduction, _ = inputs
        ctx.save_for_backward(x, y, weight)
        ctx.reduction = reduction

    @staticmethod
    def backward(ctx, ops, grad_output):
        x, y, weight = ctx.saved_tensors

        if weight is not None:
            sample_weights = weight[y]
        else:
            sample_weights = None

        dim = -1 if x.dim() > 1 else 0
        m = ops.max(x, dim=dim, keepdim=True)[0]
        x_sub_m = ops.sub(x, m)
        exp_x_sub_m = ops.exp(x_sub_m)
        sum_exp_x_sub_m = ops.sum(exp_x_sub_m, dim=dim, keepdim=True)
        log_sum_exp_x_sub_m = ops.log(sum_exp_x_sub_m)
        log_softmax = ops.sub(x_sub_m, log_sum_exp_x_sub_m)
        softmax = ops.exp(log_softmax)

        grad_x = softmax.clone()
        if grad_x.dim() == 1:
            if sample_weights is not None:
                grad_x = ops.mul(grad_x, sample_weights)
                grad_x[y] = ops.sub(grad_x[y], sample_weights)
            else:
                grad_x[y] = ops.sub(grad_x[y], ops.scalar_from_float(1.0))

        else:
            idx = torch.arange(y.size(0))
            if sample_weights is not None:
                grad_x = ops.mul(grad_x, sample_weights.view(-1, 1))
                grad_x[idx, y] = ops.sub(grad_x[idx, y], sample_weights)
            else:
                grad_x[idx, y] = ops.sub(grad_x[idx, y], ops.scalar_from_float(1.0))

        if ctx.reduction == 'mean':
            if sample_weights is not None:
                denom = ops.sum(sample_weights)
            else:
                denom = ops.scalar_from_float(y.size(0))
            grad_x = ops.div(grad_x, denom)

        elif ctx.reduction == 'none':
            if grad_x.dim() == 1:
                grad_x = ops.mul(grad_x, grad_output)
            else:
                grad_x = ops.mul(grad_x, grad_output.view(-1, 1))

        else:
            grad_x = ops.mul(grad_x, grad_output)

        return grad_x, None, None, None, None, None


import torch
from torchdt import DType

@DType.register_func(torch.zeros)
def dt_zeros(*size, out=None, dtype=None, device=None, requires_grad=False):
    if not issubclass(dtype, DType):
        raise TypeError(f"dtype must be a subclass of DType, got {dtype}")
    if isinstance(size[0], tuple):
        size = size[0]

    result = dtype(torch.zeros(size), device=device, requires_grad=requires_grad)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.zeros_like)
def dt_zeros_like(input, *, dtype=None, device=None, requires_grad=False):
    if isinstance(dtype, type) and issubclass(dtype, DType):
        return dtype(torch.zeros(input.size()), device=device, requires_grad=requires_grad)
    return torch.zeros(input.size(), dtype=dtype, device=device, requires_grad=requires_grad)

@DType.register_func(torch.ones)
def dt_ones(*size, out=None, dtype=None, device=None, requires_grad=False):
    if not issubclass(dtype, DType):
        raise TypeError(f"dtype must be a subclass of DType, got {dtype}")
    if isinstance(size[0], tuple):
        size = size[0]

    result = dtype(torch.ones(size), device=device, requires_grad=requires_grad)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.ones_like)
def dt_ones_like(input, *, dtype=None, device=None, requires_grad=False):
    if isinstance(dtype, type) and issubclass(dtype, DType):
        return dtype(torch.ones(input.size()), device=device, requires_grad=requires_grad)
    return torch.ones(input.size(), dtype=dtype, device=device, requires_grad=requires_grad)

@DType.register_func(torch.full)
def dt_full(size, fill_value, *, out=None, dtype=None, device=None, requires_grad=False):
    if not isinstance(out, DType) and not issubclass(dtype, DType):
        raise TypeError(f"dtype must be a subclass of DType, got {dtype}")
    if isinstance(size[0], tuple):
        size = size[0]

    result = dtype(torch.full(size, fill_value), device=device, requires_grad=requires_grad)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.full_like)
def dt_full_like(input, fill_value, *, dtype=None, device=None, requires_grad=False):
    if isinstance(dtype, type) and issubclass(dtype, DType):
        return dtype(torch.full(input.size(), fill_value), device=device, requires_grad=requires_grad)
    return torch.full(input.size(), fill_value, dtype=dtype, device=device, requires_grad=requires_grad)

@DType.register_func(torch.rand)
def dt_rand(*size, out=None, dtype=None, device=None, requires_grad=False):
    if not issubclass(dtype, DType):
        raise TypeError(f"dtype must be a subclass of DType, got {dtype}")
    if isinstance(size[0], tuple):
        size = size[0]

    result = dtype(torch.rand(size), device=device, requires_grad=requires_grad)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.rand_like)
def dt_rand_like(input, *, dtype=None, device=None, requires_grad=False):
    if isinstance(dtype, type) and issubclass(dtype, DType):
        return dtype(torch.rand(input.size()), device=device, requires_grad=requires_grad)
    return torch.rand(input.size(), dtype=dtype, device=device, requires_grad=requires_grad)

@DType.register_func(torch.randn)
def dt_randn(*size, out=None, dtype=None, device=None, requires_grad=False):
    if not issubclass(dtype, DType):
        raise TypeError(f"dtype must be a subclass of DType, got {dtype}")
    if isinstance(size[0], tuple):
        size = size[0]

    result = dtype(torch.randn(size), device=device, requires_grad=requires_grad)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.randn_like)
def dt_randn_like(input, *, dtype=None, device=None, requires_grad=False):
    if isinstance(dtype, type) and issubclass(dtype, DType):
        return dtype(torch.randn(input.size()), device=device, requires_grad=requires_grad)
    return torch.randn(input.size(), dtype=dtype, device=device, requires_grad=requires_grad)

@DType.register_func(torch.empty)
def dt_empty(*size, out=None, dtype=None, device=None, requires_grad=False):
    if not issubclass(dtype, DType):
        raise TypeError(f"dtype must be a subclass of DType, got {dtype}")
    if isinstance(size[0], tuple):
        size = size[0]

    result = dtype(torch.empty(size), device=device, requires_grad=requires_grad)

    if out is not None:
        return out.copy_(result)
    return result

@DType.register_func(torch.empty_like)
def dt_empty_like(input, *, dtype=None, device=None, requires_grad=False):
    if isinstance(dtype, type) and issubclass(dtype, DType):
        return dtype(torch.empty(input.size()), device=device, requires_grad=requires_grad)
    return torch.empty(input.size(), dtype=dtype, device=device, requires_grad=requires_grad)